/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

public class ColorPresentation implements PresentationType {

    private final PresentationType pt = this;

    private final class MyPresentation extends ColorModel implements Presentation {

	private final Picture p;

	public MyPresentation(Picture p) {super(31); this.p = p;}

	public MyPresentation(Picture p, String s) throws IllegalArgumentException {

	    this(p);
	    if (s.length() < 3) throw new IllegalArgumentException();

	    if (s.charAt(0) == '0') showRed.setSelected(false);
	    if (s.charAt(1) == '0') showGreen.setSelected(false);
	    if (s.charAt(2) == '0') showBlue.setSelected(false);
	}

	// ColorModel
	public int getRed(int pixel) {return pixel >> 16 & 0xff;}
	public int getGreen(int pixel) {return pixel >> 8 & 0xff;}
	public int getBlue(int pixel) {return pixel & 0xff;}
	public int getAlpha(int pixel) {return 255;}

	// Presentation
	private final JCheckBoxMenuItem showRed = new JCheckBoxMenuItem("Rotanteil", true);
	private final JCheckBoxMenuItem showGreen = new JCheckBoxMenuItem("Grnanteil", true);
	private final JCheckBoxMenuItem showBlue = new JCheckBoxMenuItem("Blauanteil", true);

	public Image getImage() throws PictureException {

	    if (p.getNumberOfLayers() != 3) 
		throw new PictureException("Es handelt sich nicht um ein Farbbild");

	    Layer red = p.getLayer(0);
	    Layer green = p.getLayer(1);
	    Layer blue = p.getLayer(2);

	    double min = 
		Math.min(red.getMinimum(), Math.min(green.getMinimum(), blue.getMinimum()));
	    double max =
		Math.max(red.getMaximum(), Math.max(green.getMaximum(), blue.getMaximum()));
	    double m = 255 / (max - min);
	    int width = p.getWidth(); int height = p.getHeight();

	    int[] id = new int[width * height];

	    for (int x = 0; x < width; x++)
		for (int y = 0; y < height; y++) {
		    if (showRed.isSelected())
			id[y * width + x] += 
			    ((((int) ((red.getValue(x, y) - min) * m)) & 0xff) << 16);
		    if (showGreen.isSelected())
			id[y * width + x] +=
			    ((((int) ((green.getValue(x, y) - min) * m)) & 0xff) << 8);
		    if (showBlue.isSelected())
			id[y * width + x] += 
			    ((((int) ((blue.getValue(x, y) - min) * m)) & 0xff) << 0);
		}
	
	    MemoryImageSource mis = new MemoryImageSource(width, height, this, id, 0, width);
	    return Toolkit.getDefaultToolkit().createImage(mis);
	}

	public JPopupMenu getJPopupMenu(ActionListener e) {
	    
	    JPopupMenu p = PresentationFactory.getJPopupMenu(this.p, e);

	    p.addSeparator();
	    p.add(showRed); showRed.addActionListener(e);
	    p.add(showGreen); showGreen.addActionListener(e);
	    p.add(showBlue); showBlue.addActionListener(e);
	    
	    return p;
	}

	public PresentationType getPresentationType() {return pt;}

	public Presentation getCopy(Picture p) {

	    if (!isCompatible(p))
		return PresentationFactory.createDefaultPresentation(p);

	    MyPresentation mp = new MyPresentation(p);
	    mp.showRed.setSelected(showRed.isSelected());
	    mp.showGreen.setSelected(showGreen.isSelected());
	    mp.showBlue.setSelected(showBlue.isSelected());
	    return mp;
	}

	public String toString() {

	    return getName() + "$" + (showRed.isSelected() ? "1" : "0") + 
		(showGreen.isSelected() ? "1" : "0") + (showBlue.isSelected() ? "1" : "0");
	}
    }

    public boolean isCompatible(Picture mlp) {

	return (mlp.getNumberOfLayers() == 3);
    }

    public String getName() {return "Farb-Darstellung";}

    public Presentation createPresentation(Picture p) throws PictureException {

	if (!isCompatible(p))
	    throw new PictureException("Das Bild kann nicht als Farbbild dargestellt werden");

	return new MyPresentation(p);
    }

    public Presentation createPresentation(Picture p, String s) 
	throws PictureException, IllegalArgumentException {

	if (!isCompatible(p))
	    throw new PictureException("Das Bild kann nicht als Farbbild dargestellt werden");

	return new MyPresentation(p, s);
    }
}
