/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture;
import picture.math.Term;
import picture.filter.*;

/**
 * Diese Klasse implementiert einen Filter. Sie wird von der Klasse PassFilter zur Berechnung
 * von Hoch-, Tief- und Bandpassfiltern verwendet.
 */
public class FilterPicture extends Picture {

    /** Kennzeichnet einen Bandpassfilter */
    public static final int BANDPASSFILTER = PassFilter.BANDPASSFILTER;

    /** Kennzeichnet einen Hochpassfilter */
    public static final int HIGHPASSFILTER = PassFilter.HIGHPASSFILTER;
    
    /** Kennzeichnet einen Tiefpassfilter */
    public static final int LOWPASSFILTER = PassFilter.LOWPASSFILTER;

    private boolean complex = false;

    public FilterPicture() {}

    /**
     * Erzeugt ein FilterPicture aus dem angegebenen Model. Intern wird mit dem Ursprung
     * oben links in der Ecke gearbeitet. Die graphische Benutzerschnittstelle zeigt 
     * aber aus Anschauungsgruenden den Ursprung in der Mitte. Wenn der Benutzer eine
     * Datei als Filter auswaehlt, wird diese wohl dann auch den Ursprung in der Mitte
     * haben. Dieser Konstruktor erzeugt aus einem solchen Bild eines, in dem der Ursprung
     * oben links liegt, indem einfach das untere Viertel des uebergebenen Bildes betrachtet
     * wird. Beachte: Dazu muss das uebergebene Bild punktsymmetrisch zum Mittelpunkt sein.
     */
    public FilterPicture(Picture pm) {
	
	width = pm.getWidth();
	height = pm.getHeight();
	int depth = pm.getNumberOfLayers();
	
	double[][][] data = new double[depth][width][height];
	
	int w2 = width / 2;
	int h2 = height / 2;
	
	double[] c;
	
	for (int x = w2; x < width; x++)
	    for (int y = h2; y < height; y++) {
		c = pm.getValue(x, y);
		for (int z = 0; z < depth; z++) {
		    data[z][(x - w2) * 2][(y - h2) * 2] = c[z];
		    data[z][(x - w2) * 2][(y - h2) * 2 + 1] = c[z];
		    data[z][(x - w2) * 2 + 1][(y - h2) * 2] = c[z];
		    data[z][(x - w2) * 2 + 1][(y - h2) * 2 + 1] = c[z];
		}
	    }

	layers = new Layer[depth];
	for (int i = 0; i < depth; i++)
	    layers[i] = new Layer(data[i]);	

	complex = (depth == 2);
    }
    
    /**
     * Macht die Transformation des Konstruktors FilterPicture(PictureModel) wieder
     * rueckgaengig.
     */
    private FilterPicture(int width, int height, Layer[] pics) {
	    
	this.width = width;
	this.height = height;
	
	double[][][] data = new double[pics.length][width][height];
	
	int w2 = width / 2; int w2m1 = w2 - 1;
	int h2 = height / 2; int h2m1 = h2 - 1;

	double c;
	for (int x = 0; x < width; x += 2)
	    for (int y = 0; y < height; y += 2) 
		for (int z = 0; z < pics.length; z++) {
		    c = pics[z].getValue(x, y);
		    data[z][w2 + x / 2][h2 + y / 2] = c;
		    data[z][w2m1 - x / 2][h2 + y / 2] = c;
		    data[z][w2 + x / 2][h2m1 - y / 2] = c;
		    data[z][w2m1 - x / 2][h2m1 - y / 2] = c;
		}

	this.layers = new Layer[pics.length];
	for (int i = 0; i < pics.length; i++)
	    this.layers[i] = new Layer(data[i]);

	complex = (pics.length == 2);
    }
    
    /**
     * Erzeugt einen neuen Bandpassfilter. Das Bild ist komplett schwarz (0), ausser
     * im filterspezifischen Bereich: Fuer den Bandpassfilter werden alle Punkte,
     * deren Abstand zum Ursprung (0, 0) zwischen <tt>min</tt> und <tt>max</tt> liegt,
     * auf weiss (1) gesetzt. Fuer den Hochpassfilter werden alle Punkte, deren
     * Abstand zum Mittelpunkt groesser ist als <tt>min</tt>, auf weiss gesetzt und
     * fuer den Tiefpassfilter werden alle Punkte, deren Abstand zum Mittelpunkt
     * kleiner ist als <tt>max</tt>, auf weiss gesetzt. min und max seien auf (0..1)
     * normiert.
     * <p>
     * Wir ein Gradueller Filter gewuenscht, so werden zur Berechnung die dem angegebenen
     * Filter entsprechenden Methoden dieser Klasse verwendet.
     * 
     * @param mode: BANDPASSFILTER, HIGHPASSFILTER oder LOWPASSFILTER.
     * @param min: Innenring 
     * @param max: Aussenring
     * @param width: Breite des zu erzeugenden Bildes
     * @param height: Hoehe des zu erzeugenden Bildes
     * @param graduell: <tt>true</tt>, um einen graduellen Filter zu erzeugen
     */
    public FilterPicture(int mode, double min, double max, 
			 int width, int height, boolean graduell) {
	
	this.width = width;
	this.height = height;
	
	double[][] data = new double[width][height];
	
	double weiss = 1; double schwarz = 0;
	
	if (!graduell)
	    for (int i = 0; i < width; i++)
		for (int j = 0; j < height; j++)
		    data[i][j] = schwarz;
	
	min *= Math.min(width, height);
	max *= Math.min(width, height);
	
	int w2 = width / 2;
	int h2 = height / 2;
	
	double norm;
	switch (mode) {
	case BANDPASSFILTER: 
	    if (graduell)
		for (int i = 0; i < width; i++)
		    for (int j = 0; j < height; j++)
			data[i][j] = bandpass(norm(i, j), min, max);
	    else
		for (int i = 0; i < width; i++) 
		    for (int j = 0; j < height; j++) {
			norm = norm(i, j);
			if (norm >= min && norm <= max)
			    data[i][j] = weiss;
		    }		    
	    break;
	case HIGHPASSFILTER:
	    if (graduell)
		for (int i = 0; i < width; i++)
		    for (int j = 0; j < height; j++)
			data[i][j] = highpass(norm(i, j), min);
	    else
		for (int i = 0; i < width; i++)
		    for (int j = 0; j < height; j++) {
			norm = norm(i, j);
			if (norm >= min)
			    data[i][j] = weiss;
		    }
	    break;
	case LOWPASSFILTER:
	    if (graduell)
		for (int i = 0; i < width; i++)
		    for (int j = 0; j < height; j++)
			data[i][j] = lowpass(norm(i, j), max);
	    else
		for (int i = 0; i < width; i++)
		    for (int j = 0; j < height; j++) {
			norm = norm(i, j);
			if (norm <= max)
			    data[i][j] = weiss;
		    }
	    break;       
	}		

	layers = new Layer[]{new Layer(data)};
    }
    
    /**
     * Erzeugt eine neue Filterdatei. Die Funktion <tt>f</tt> wird dabei als Filterfunktion
     * verwendet, wobei nur der Realteil der Funktionswerte ber&uuml;cksichtigt wird.     
     *
     * @exception PictureException, falls f ausser x noch weitere Variablen enthaelt.
     */
    public FilterPicture(Term f, int width, int height) throws PictureException {
	
	this.width = width; 
	this.height = height;
	
	double[][] data = new double[width][height];
	
	double scale = Math.sqrt(2) / Math.max(width, height);
	int w2 = width / 2;
	int h2 = height / 2;
	
	String [] x = new String[]{"x"};
	Complex [] z = new Complex[1];
	
	for (int i = 0; i < width; i++)
	    for (int j = 0; j < height; j++) {
		z[0] = new Complex(norm(i, j) * scale, 0);
		z[0] = f.einsetzen(x, z);
		if (z[0] == null)
		    throw new PictureException("Die Funktion darf nur von x abhngen.");
		data[i][j] = z[0].getRealPart();
	    }

	layers = new Layer[]{new Layer(data)};
    }
    
    /**
     * Intern wird der Mittelpunkt links oben angenommen. Diese Methode verschiebt den
     * Punkt links oben in die Mitte des Bildes und setzt den Rest derart fort, dass
     * ein zum Mittelpunkt punktsymmetrisches Bild entsteht, das den Anzeigegepflogenheiten
     * in <tt>FilterPane</tt> entspricht.
     * Achtung: Dabei gehen Informationen verloren, da das eigentliche Bild auf 25% 
     * zusammengestaucht wird.<p>
     * Das Datenformat des zurueckgegebenen Bildes ist pgm.
     */
    public Picture getViewableInstance() {
	
	return new FilterPicture(width, height, layers);
    }

    public boolean isComplex() {return complex;}

    // #############################################################################
    // ###  Berechnungsmethoden  ###################################################
    // #############################################################################

    private static double norm(double d, double e) {return Math.sqrt(d*d + e*e);}
    private static double sqr(double d) {return d * d;}
    private static final double log2 = Math.log(2);

    /**
     * Berechnet die fuer einen graduellen Tiefpass-Filter verwendete Funktion
     * <center>
     *  <i>f</i>: [0, 1] -> [0, 1]
     * </center>
     * an der Stelle <i>x</i>.<p>
     * Mit <i>x</i><sub>0</sub> wird Grenzstelle angegeben. Wir definieren
     * <center>
     *  <i>t</i> := min(<i>x</i><sub>0</sub>, 1 - <i>x</i><sub>0</sub>).
     * </center>
     * Fuer die Funktion <i>f</i> gilt dann:
     * <ul>
     *  <li><i>f</i>(<i>x</i>) = 1 (<i>x</i> &lt; <i>x</i><sub>0</sub> - <i>t</i>)</li>
     *  <li><i>f</i>(<i>x</i><sub>0</sub>) = 1/2</li>
     *  <li><i>f</i>(<i>x</i>) = 0 (<i>x</i> &gt; <i>x</i><sub>0</sub> + <i>t</i>)</li>
     *  <li><i>f</i> ist stetig differenzierbar
     * </ul>
     * Zur Berechnung von werden Exponentialfunktionen herangezogen.
     * 
     * @param x Stelle, an der die Funktion ausgewertet werden soll
     * @param x0 Grenzstelle.
     * @return <i>f</i>(<i>x</i>)
     */
    public static double lowpass(double x, double x0) {

	double t = Math.min(x0, 1 - x0);
	double x1 = x0 - t;
	double x2 = x0 + t;

	if (x <= x1)	  t = 1;
	else if (x >= x2) t = 0;
	else if (x <= x0) t = Math.exp(-log2 * sqr((x1 - x) / (x1 - x0)));
	else              t = 1 - Math.exp(-log2 * sqr((x2 - x) / (x2 - x0)));

	return t;
    }

    /**
     * Berechnet die fuer einen graduellen Hochpassfilter verwendete Funktion 
     * <center>
     *  <i>h</i>: [0, 1] -> [0, 1]
     * </center>
     * an der Stelle <i>x</i>. Es gilt
     * <center>
     *  <i>h</i>(<i>x</i>) := 1 - <i>f</i>(<i>x</i>),
     * </center>
     * wobei <i>f</i>: [0, 1] -> [0, 1] die Hochpassfilterfunktion ist. Naehere Informationen
     * siehe also dort.
     * @param x Stelle, an der die Funktion ausgewertet werden soll.
     * @param x0 Grenzstelle
     * @see #highpass
     * @return <i>h</i>(<i>x</i>)
     */
    public static double highpass(double x, double x0) {

	return 1 - lowpass(x, x0);
    }

    /**
     * Berechnet die fuer graduelle Bandpassfilter verwendete Funktion 
     * <center>
     *  <i>b</i>: [0, 1] -> [0, 1]
     * </center>
     * an der Stelle <i>x</i>. Mit <i>x</i><sub>1</sub> und <i>x</i><sub>2</sub> werden
     * die Grenzstellen spezifiziert. Fuer die Funktion <i>b</i> gilt:
     * <ul>
     *  <li><i>b</i>((<i>x</i><sub>1</sub> + <i>x</i><sub>2</sub>) / 2) = 1</li>
     *  <li><i>b</i>(<i>x</i><sub><i>i</i></sub>) = 1/2 (<i>i</i> = 1, 2)</li>
     *  <li><i>b</i> ist stetig differenzierbar.
     * </ul>
     * 
     * @param x Stelle, an der die Funktion ausgewertet werden soll.
     * @param x0 erste Grenzstelle
     * @param x1 zweite Grenzstelle
     * @return <i>b</i>(<i>x</i>)
     */
    public static double bandpass(double x, double x0, double x1) {

	double x2 = (x0 + x1) / 2;
	return Math.exp(-log2 * sqr((x - x2) / (x2 - x1)));
    }
}


    
