/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture;

/**
 * Implementiert die vier Gundrechenarten +, *, -, /, mit denen zwei Bilder
 * pixelweise verknuepft werden koennen.
 */
public class FundamentalOperation extends Combination {

    public static final int SUM = 1;
    public static final int DIFFERENCE = 2;
    public static final int PRODUCT = 3;
    public static final int QUOTIENT = 4;
    public static final int COMPLEX_PRODUCT = 5;

    private int art = PRODUCT;

    /**
     * Erzeugt die angegebene Grundrechenart. Bei Verwendung dieses
     * Konstruktors wirft die Methode of(Layer) eine null-Pointer-Exception.
     * @param fundamentalOperation. Eins von SUM, DIFFERENCE, PRODUCT, QUOTIENT.
     */
    public FundamentalOperation(int fundamentalOperation) {art = fundamentalOperation;}

    public int getNumberOfProgressSteps(Picture p) {

	return p.getWidth() * ((art == COMPLEX_PRODUCT) ? 1 : p.getNumberOfLayers());
    }

    /**
     * Verknuepft die Pictureobjekte <tt>a</tt> und <tt>b</tt> mit der
     * im Konstruktor angegebenen Verknuepfung.
     * @param a: erster Operand
     * @param b: zweiter Operand
     * @return <tt>a</tt> # <tt>b</tt>, wobei # fuer die im Konstruktor angegebene
     * Verknuepfung steht.
     * @exception PictureException, falls die Groesse der Bilder <tt>a</tt> und <tt>b</tt>
     * nicht uebereinstimmt.
     */
    public Layer of(ProgressHandler ph, Layer a, Layer b) throws PictureException {

	int width = a.getWidth();
	int height = a.getHeight();

	if (width != b.getWidth() || height != b.getHeight())
	    throw new PictureException("die beiden zu verknpfenden Bilder\n" +
				       "passen in der Gre nicht zusammen.");

	double[][] data = new double[width][height];
	double a0; double b0;

	for (int i = 0; i < width; i++) {
	    if (ph != null) ph.step();
	    for (int j = 0; j < height; j++) {
		a0 = a.getValue(i, j); b0 = b.getValue(i, j);
		switch (art) {
		case SUM:        data[i][j] = a0 + b0; break;
		case DIFFERENCE: data[i][j] = a0 - b0; break;
		case PRODUCT:    data[i][j] = a0 * b0; break;
		case QUOTIENT:   data[i][i] = (b0 == 0) ? 0 : a0 / b0; break;
		case COMPLEX_PRODUCT: 
		    throw new PictureException("Komplexes Produkt ist nicht auf einzelnen\n" +
					       "Layern mglich.");
		default: 
		    throw new PictureException("Unbekannte Verknpfung.");
		}
	    }
	}
				
	return createNewInstance(a, data);
    }

    public Picture of(Picture a, Picture b) 
	throws PictureException {

	if (art == COMPLEX_PRODUCT) {
	    
	    if (a.getNumberOfLayers() == 1) a = realToComplex(a);
	    if (b.getNumberOfLayers() == 1) b = realToComplex(b);
	    
	}

	int la = a.getNumberOfLayers();
	if (a.getWidth() != b.getWidth() || a.getHeight() != b.getHeight() || 
	    la != b.getNumberOfLayers())
 
	    throw new PictureException("Die Bilder haben nicht die gleiche Gre oder\n" +
				       "eine unterschiedliche Anzahl an Ebenen.");

	Picture erg = null;
	if (art == COMPLEX_PRODUCT) {

	    if (!a.isComplex() || !b.isComplex())
		throw new PictureException("Komplexes Produkt ist nur auf komplexen Bilder\n" +
					   "mglich.");

	    double[][] re = new double[a.getWidth()][a.getHeight()];
	    double[][] im = new double[a.getWidth()][a.getHeight()];

	    for (int x = 0; x < a.getWidth(); x++) {
		a.fireProgressEvent();
		for (int y = 0; y < a.getHeight(); y++) {

		    Complex c = new Complex(a.getValue(x, y)).mul(new Complex(b.getValue(x, y)));
		    re[x][y] = c.getRealPart();
		    im[x][y] = c.getImaginaryPart();
		}
	    }

	    erg = new Picture(a, new Layer[]{new Layer(re), new Layer(im)});
		    
	} else {

	    Layer [] p = new Layer[la];
	    for (int i = 0; i < la; i++) 
		p[i] = of(a.getLayer(i), b.getLayer(i));

	    erg = new Picture(a, p);
	}

	return erg;
    }

    private Picture realToComplex(Picture p) {

	double[][] d = new double[p.getWidth()][p.getHeight()];

	Layer[] pics = new Layer[]{p.getLayer(0), new Layer(d)};

	return new Picture(p, pics) {
	    
	    public boolean isComplex() {return true;}
	};
    }
}
