/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.filetype;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import javax.swing.JFrame;

import picture.io.*;
import picture.gui.*;
import picture.filter.*;
import picture.*;

/**
 * PictureType-Unterklasse fr JPF (Jipp Picture Format) - Dateien.
 *
 * @version 0.2
 * @author Stefan Rhrich 
 */
public class JPFPictureType extends PictureType {

    static {
	PictureTypeFactory.registerPictureType(new JPFPictureType());

	/* nur damit die kompiliert werden ... */
	new DefaultStreams();
	new RLEOutputStream(null);
	new RLEInputStream(null);
	new DiffOutputStream(null);
	new DiffInputStream(null);
    }

    /**
     * Dummy zum Aufrufen des Static-Initializers ...
     */
    public static final void init() {
    }

    public String getExtension() {
	return "jpf";
    }

    public String getDescription() {
	return "Jipp Picture Format-Bild";
    }

    public DialogPane getSaveOptionsPane(JFrame f, Picture p) throws PictureException {
	return new JippSavePane(f, p);
    }

    private static class JPFHeaderCommand {
	String command;
	String[] args;
    }
    
    private static class RefArgInt {
	int i = 0;
    }

    private static final int MY_FILE_VERSION = 1;

    private static final String S_VERSION = "version";
    private static final String S_LAYERS = "layers";
    private static final String S_SCALE = "intensityScale";
    private static final String S_DIMENSION = "dimension";
    private static final String S_MAX_GRAY = "maxGrayValue";
    private static final String S_DATA_STREAM = "dataStream";
    private static final String S_END_HEADER = "endHeader";
    private static final String S_CHECKSUM = "dataChecksum";
    private static final String S_WAVELET = "wavelet";
    private static final String S_SCALE_DETAILS = "scaleDetails";
    private static final String S_MINMAX = "minimumMaximum";
    private static final String S_PRESENTATION = "presentation";

    public Picture load(InputStream origin) throws PictureException, IOException {
	InputStream in = origin;
	boolean compressHeader;

	int width = 0, height = 0;

	String md5 = null;

	char c;

	if((char)in.read() == 'J' && (char)in.read() == 'I' &&
	   (char)in.read() == 'P' && ((c = (char)in.read()) == 'Z' || (c == 'P'))
	   && (char)in.read() == '\n')
	    compressHeader = (c == 'Z');
	else
	    throw new PictureException("Fehler: Keine JPF-Datei!");

	if(compressHeader)
	    in = new BWTInputStream(new MTFInputStream(new ArithInputStream(new KnownLengthInputStream(in))));

	JPFHeaderCommand hc;
	Vector comments = new Vector();

	double[] loff = new double[1];
	double[] lscale = new double[1];
	double[] lmin = new double[1];
	double[] lmax = new double[1];

	InputStream newin = origin;

	Class is;
	try {
	    is = Class.forName("java.io.InputStream");
	} catch(ClassNotFoundException e) {
	    GUIHelp.error(e);
	    throw new PictureException("JPF: Fehler bei Ladeinitialisierung!");
	}

	boolean waveComp = false, waveScale = true;
	int waveOhneDetailsIter = 0, waveMitDetailsIter;
	int scale = 1;
	long numberOfGrayScales = 255;
	int version = -1, lc = 1;

	String spres = "";

	Wavelet wod = null, wmd = null;

	while((hc = readCommand(in, comments)) != null) {
	    if(hc.command.equals(S_VERSION)) {
		version = Integer.parseInt(hc.args[0]);
		if(version != MY_FILE_VERSION)
		    throw new PictureException("JPF: Falsche Dateiversion!");
		continue;
	    }
	    if(hc.command.equals(S_SCALE)) {
		int slayer = Integer.parseInt(hc.args[0]);
		loff[slayer] = Double.parseDouble(hc.args[1]);
		lscale[slayer] = Double.parseDouble(hc.args[2]);
		continue;
	    }
	    if(hc.command.equals(S_DIMENSION)) {
		width = Integer.parseInt(hc.args[0]);
		height = Integer.parseInt(hc.args[1]);
		if(hc.args[2] != null) {
		    lc = Integer.parseInt(hc.args[2]);
		    if(lc != 1) {
			loff = new double[lc];
			lscale = new double[lc];
			lmin = new double[lc];
			lmax = new double[lc];
		    }
		}
		continue;
	    }
	    if(hc.command.equals(S_MAX_GRAY)) {
		numberOfGrayScales = Long.parseLong(hc.args[0]);
		continue;
	    }
	    if(hc.command.equals(S_DATA_STREAM)) {
		try {
		    newin = (InputStream)Class.forName(hc.args[0] + "InputStream").getConstructor(new Class[] {is}).newInstance(new Object[] {newin});
		} catch(ClassNotFoundException e) {
		    GUIHelp.error(e);
		} catch(NoSuchMethodException e) {
		    GUIHelp.error(e);
		} catch(InstantiationException e) {
		    GUIHelp.error(e);
		} catch(IllegalAccessException e) {
		    GUIHelp.error(e);
		} catch(java.lang.reflect.InvocationTargetException e) {
		    GUIHelp.error(e);
		}
		continue;
	    }
	    if(hc.command.equals(S_CHECKSUM)) {
		if(hc.args[0].equals("MD5"))
		    md5 = hc.args[1];
		continue;
	    }
	    if(hc.command.equals(S_WAVELET)) {
		if(hc.args[0].equals("Daubechies")) {
		    int n = Integer.parseInt(hc.args[1]);
		    waveScale = hc.args[2].equals(S_SCALE_DETAILS);
		    waveMitDetailsIter = Integer.parseInt(hc.args[3]);
		    waveOhneDetailsIter = Integer.parseInt(hc.args[4]);
		    wod = Wavelet.getDaubechiesWavelet(n, Wavelet.SYNTHESIS_WITHOUT_DETAILS, waveScale);
		    wmd = Wavelet.getDaubechiesWavelet(n, Wavelet.MULTI_SYNTHESIS, waveScale);
		    wmd.setMultiSteps(waveMitDetailsIter);
		} else
		    if(hc.args[0].equals("Haar")) {
			waveScale = hc.args[1].equals(S_SCALE_DETAILS);
			waveMitDetailsIter = Integer.parseInt(hc.args[2]);
			waveOhneDetailsIter = Integer.parseInt(hc.args[3]);
			wod = Wavelet.getHaarWavelet(Wavelet.SYNTHESIS_WITHOUT_DETAILS, waveScale);
			wmd = Wavelet.getHaarWavelet(Wavelet.MULTI_SYNTHESIS, waveScale);
			wmd.setMultiSteps(waveMitDetailsIter);
		    }

		waveComp = true;

		for(int i = 0; i < waveOhneDetailsIter; i++)
		    scale *= 2;

		continue;
	    }
	    if(hc.command.equals(S_MINMAX)) {
		int slayer = Integer.parseInt(hc.args[0]);
		lmin[slayer] = Double.parseDouble(hc.args[1]);
		lmax[slayer] = Double.parseDouble(hc.args[2]);
		continue;
	    }
	    if(hc.command.equals(S_PRESENTATION)) {
		for(int i = 0; i < hc.args.length; i++)
		    spres += hc.args[i] + " ";
		continue;
	    }
	}

	//	System.err.println("Header gelesen: " + width + " " + height + " " + saveComplex);

	int pw;
	if(numberOfGrayScales == -1)
	    pw = 8;
	else
	    pw = (int)Math.round(Math.log(numberOfGrayScales)/(Math.log(2)*8.0)+0.5);
	int sw = width, sh = height;
	width = width/scale;
	height = height/scale;
	int size = lc*width*height*pw;
	byte[] bdat = new byte[size];
	RefArgInt pos = new RefArgInt();

	if(newin.read(bdat) != size)
	    throw new PictureException("JPF: Fehler beim Lesen der Daten!");

	//	System.err.println("Daten gelesen: " + size);

	if(md5 != null)
	    if(!md5.equals(MD5.md5(bdat)))
		throw new PictureException("Fehler bei Datenprfsumme: " + md5 + " != " + MD5.md5(bdat));
	
	newin.close();

	double[][][] data = new double[lc][width][height];

	DataInputStream datin = new DataInputStream(new ByteArrayInputStream(bdat));

	if(numberOfGrayScales == -1)
	    for(int i = 0; i < lc; i++)
		for(int y = 0; y < height; y++)
		    for(int x = 0; x < width; x++)
			data[i][x][y] = datin.readDouble();
	else
	    for(int i = 0; i < lc; i++)
		for(int y = 0; y < height; y++)
		    for(int x = 0; x < width; x++)
			data[i][x][y] = readPixel(bdat, pos, pw, loff[i], lscale[i]);

	Layer[] layers = new Layer[lc];

	for(int i = 0; i < lc; i++)
	    layers[i] = new Layer(data[i]);

	Picture p = new Picture(layers);

	if(spres.equals("")) {
	    if(lc == 3)
		p.setPresentation(PresentationFactory.presentationByName(p, "Farb-Darstellung"));
	    if(lc == 2)
		p.setPresentation(PresentationFactory.presentationByName(p, "Komplexe Darstellung"));
	}

	if (waveComp) {
	    p = wmd.of(p);
	    for (int i = 0; i < waveOhneDetailsIter; i++)
		p = wod.of(p);
	}

	for(int i = 0; i < lc; i++) {
	    p.getLayer(i).setMinimum(lmin[i]);
	    p.getLayer(i).setMaximum(lmax[i]);
	}

	p.setPresentation(PresentationFactory.presentationByName(p, spres));

	return p;
    }

    private static final JPFHeaderCommand readCommand(InputStream in, Vector comments) throws IOException {
	//	System.err.print("readCommand: ");
	int t = 0;
	JPFHeaderCommand hc = new JPFHeaderCommand();
	boolean fertig = false;
	boolean first = true;
	boolean comment = false;
	StringBuffer s = new StringBuffer();

	while(!fertig && (t = in.read()) != -1) {
	    if(t == '\n') {
		first = true;
		if(comment) {
		    comment = false;
		    comments.add(s.toString());
		    s = new StringBuffer();
		} else
		    fertig = (s.length() > 0);
	    } else
		if(first && t == '#')
		    comment = true;
		else
		    s.append((char)t);
	}

	//	System.err.println(s);

	if(t == -1 || s.length() == 0)
	    return null;

	StringTokenizer st = new StringTokenizer(s.toString());

	if(st.hasMoreTokens())
	    hc.command = st.nextToken();
	else
	    return null;

	if(hc.command.equals(S_END_HEADER))
	    return null;

	Vector argv = new Vector();

	while(st.hasMoreTokens())
	    argv.add(st.nextToken());

	hc.args = vectorToStringArray(argv);

	return hc;
    }

    public void save(Picture p, PictureProperty pp, OutputStream origout) throws PictureException, IOException {

	OutputStream out = new BufferedOutputStream(origout);

	boolean waveComp = false, waveScale = true;
	int waveOhneDetailsIter = 0, waveMitDetailsIter = 0;
	Enumeration streams = null;
	String[] comments = null;
	String checksum = null;
	boolean compressHeader = false;
	boolean digitize = true;
	long mg = 255;
	int pres;

	int lc = p.getNumberOfLayers();

	Wavelet wod, wmd;

	if(pp instanceof JippPictureProperty) {
	    JippPictureProperty jpp = (JippPictureProperty)pp;
	    waveComp = jpp.waveletCompression();
	    waveMitDetailsIter = jpp.getTotalNumberOfIterationsInWaveletAnalysis()
		- jpp.getNumberOfIterationsWithoutDetailsInWaveletAnalysis();
	    waveOhneDetailsIter = jpp.getNumberOfIterationsWithoutDetailsInWaveletAnalysis();
	    waveScale = jpp.waveletDetailsScaling();
	    streams = jpp.getCodingStreams();
	    comments = jpp.getComment();
	    checksum = jpp.getChecksum();
	    compressHeader = jpp.compressHeader();
	    digitize = jpp.digitize();
	    mg = jpp.getNumberOfGrayScales();
	}

	if(!digitize)
	    mg = -1;

	writeLine(out, compressHeader ? "JIPZ" : "JIPP");

	out.flush();

	if(compressHeader)
	    out = new BufferedOutputStream(new BWTOutputStream(new MTFOutputStream(new ArithOutputStream(new KnownLengthOutputStream(out)))), 32768);

	if(comments != null)
	    for(int i = 0; i < comments.length; i++)
		writeLine(out, "# " + comments[i]);

	writeLine(out, S_VERSION + " " + MY_FILE_VERSION);

	writeLine(out, S_DIMENSION + " " + p.getWidth() + " " + p.getHeight() + " " + p.getNumberOfLayers());

	for(int i = 0; i < lc; i++)
	    writeLine(out, S_MINMAX + " " + i + " " + p.getLayer(i).getMinimum() + " " + p.getLayer(i).getMaximum());

	writeLine(out, S_PRESENTATION + " " + p.getPresentation().toString());

	if (waveComp) {

	    wod = Wavelet.getDaubechiesWavelet(6, Wavelet.ANALYSIS_WITHOUT_DETAILS, waveScale);
	    wmd = Wavelet.getDaubechiesWavelet(6, Wavelet.MULTI_ANALYSIS, waveScale);
	    wmd.setMultiSteps(waveMitDetailsIter);
	    writeLine(out, S_WAVELET + " Daubechies 6 " + 
		      (waveScale?S_SCALE_DETAILS:"smallDetails") + " " +
		      waveMitDetailsIter + " " + waveOhneDetailsIter);

	    for (int i = 0; i < waveOhneDetailsIter; i++)
		p = wod.of(p);

	    p = wmd.of(p);
	}

	double[] lmin = new double[lc];
	double[] lmax = new double[lc];

	for(int i = 0; i < lc; i++) {
	    lmin[i] = p.getLayer(i).getMinimum();
	    lmax[i] = p.getLayer(i).getMaximum();
	}

	double[] lm = new double[lc];

	writeLine(out, S_MAX_GRAY + " " + mg);

	if(digitize) {
	    for(int i = 0; i < lc; i++)
		lm[i] = (lmax[i]-lmin[i])/mg;

	    for(int i = 0; i < lc; i++)
		writeLine(out, S_SCALE + " " + i + " " + lmin[i] + " " + lm[i]);
	}

	OutputStream newout = new BufferedOutputStream(origout);

	if(streams != null) {
	    Class os;
	    try {
		os = Class.forName("java.io.OutputStream");
	    } catch(ClassNotFoundException e) {
		GUIHelp.error(e);
		return;
	    }
	    while(streams.hasMoreElements()) {
		String s = (String)streams.nextElement();
		try {
		    // Reflection-API at it's best ...
		    newout = (OutputStream)Class.forName(s + "OutputStream").getConstructor(new Class[] {os}).newInstance(new Object[] {newout});
		    writeLine(out, S_DATA_STREAM + " " + s);
		} catch(ClassNotFoundException e) {
		    GUIHelp.error(e);
		} catch(NoSuchMethodException e) {
		    GUIHelp.error(e);
		} catch(InstantiationException e) {
		    GUIHelp.error(e);
		} catch(IllegalAccessException e) {
		    GUIHelp.error(e);
		} catch(java.lang.reflect.InvocationTargetException e) {
		    GUIHelp.error(e);
		}
	    }
	}

	byte[] b = convertData2Bytes(p, lmin, lmax, mg);

	if(checksum != null && checksum.equals("MD5"))
	    writeLine(out, S_CHECKSUM + " MD5 " + MD5.md5(b));

	writeLine(out, S_END_HEADER);
	out.flush();

	out = newout;
	
	out.write(b);

	out.close();
    }

    /**
     * Writes a string to specified OutputStream.
     *
     * @param out OutputStream to write to.
     * @param s String to be written.
     * @exception IOException Problems on writing to OutputStream
     */
    private static final void writeString(OutputStream out, String s) throws IOException {
        out.write(s.getBytes());
    }

    private static final void writeLine(OutputStream out, String s) throws IOException {
        out.write((s+"\n").getBytes());
    }

    private static final double readPixel(byte[] buf, RefArgInt pos, int pw, double offset, double scale)
	throws PictureException {
	long v = 0;

	if(pw <= 0 || pw > 8)
	    throw new PictureException("JPF: Ungltige Pixelbreite!");

	// keine breaks beabsichtigt!
	switch(pw) {
	case 8:
	    v += ByteUtils.unsignedByte2Int(buf[pos.i++]) << 56;
	case 7:
	    v += ByteUtils.unsignedByte2Int(buf[pos.i++]) << 48;
	case 6:
	    v += ByteUtils.unsignedByte2Int(buf[pos.i++]) << 40;
	case 5:
	    v += ByteUtils.unsignedByte2Int(buf[pos.i++]) << 32;
	case 4:
	    v += ByteUtils.unsignedByte2Int(buf[pos.i++]) << 24;
	case 3:
	    v += ByteUtils.unsignedByte2Int(buf[pos.i++]) << 16;
	case 2:
	    v += ByteUtils.unsignedByte2Int(buf[pos.i++]) << 8;
	case 1:
	    v += ByteUtils.unsignedByte2Int(buf[pos.i++]);
	}

	return v*scale+offset;
    }

    private static final int savePixel(byte[] buf, int pos, long p, int pw) {
	// keine breaks beabsichtigt!
	switch(pw) {
	case 8:
	    buf[pos++] = ((byte)((p >> 56) & 0xff));
	case 7:
	    buf[pos++] = ((byte)((p >> 48) & 0xff));
	case 6:
	    buf[pos++] = ((byte)((p >> 40) & 0xff));
	case 5:
	    buf[pos++] = ((byte)((p >> 32) & 0xff));
	case 4:
	    buf[pos++] = ((byte)((p >> 24) & 0xff));
	case 3:
	    buf[pos++] = ((byte)((p >> 16) & 0xff));
	case 2:
	    buf[pos++] = ((byte)((p >> 8)& 0xff));
	case 1:
	    buf[pos++] = ((byte)(p & 0xff));
	}

	return pos;
    }

    private static final long digitizePixel(double d, long depth, double min, double max) {
        double m = depth / (max - min);

        if (d > max)
            return depth;
        else if (d < min)
            return 0;

        return (long)((d - min) * m);
    }

    private static byte[] convertData2Bytes(Picture p, double[] lmin, double[] lmax, long mg)
	throws IOException {
	byte[] bdatbuf;
	int pos = 0;
	int lc = p.getNumberOfLayers();

	if(mg == -1) {
	    ByteArrayOutputStream bout = new ByteArrayOutputStream(lc*p.getWidth()*p.getHeight()*8);
	    DataOutputStream out = new DataOutputStream(bout);

	    for(int i = 0; i < lc; i++) {
		Layer l = p.getLayer(i);
		for(int y = 0; y < p.getHeight(); y++)
		    for(int x = 0; x < p.getWidth(); x++)
			out.writeDouble(l.data[x][y]);
	    }

	    out.flush();
	    bdatbuf = bout.toByteArray();
	} else {
	    int pw = (int)Math.round(Math.log(mg)/(Math.log(2)*8.0)+0.5);
	    bdatbuf = new byte[lc*p.getWidth()*p.getHeight()*pw];

	    for(int i = 0; i < lc; i++) {
		Layer l = p.getLayer(i);
		for(int y = 0; y < p.getHeight(); y++)
		    for(int x = 0; x < p.getWidth(); x++)
			pos = savePixel(bdatbuf, pos, digitizePixel(l.data[x][y], mg, lmin[i], lmax[i]), pw);
	    }
	}
	return bdatbuf;
    }

    /**
     * Converts a vector of strings to String[].
     *
     * @param v Vector to convert.
     * @return generated String array.
     */
    protected static String[] vectorToStringArray(Vector v) {

        String[] s = new String[v.size()];

        for (int i = 0; i < v.size(); i++)
            s[i] = (String)v.elementAt(i);

        return s;
    }
}
