/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.filter;

import picture.*;
import picture.math.Term;

/**
 * Klasse zur Berechnung der schnellen Fouriertransformation.  <p> Zur
 * Vereinfachung wird die folgende Vereinbarung getroffen: Die FFT
 * selbst wird im Folgenden als <i>Vorwaerts</i>-FFT, die inverse FFT
 * als <i>Rueckwaerts</i>-FFT bezeichnet.
 *
 * @author Stefan Roehrich
 * @author Manuel Kauers
 * @version 1.1
 */
public class FFT extends DefaultTransformation {

    /**
     * Liefert i, falls powerOfTwo = 2^i.
     */
    protected final static int numberOfBitsNeeded(int powerOfTwo) {

	if( powerOfTwo < 2 )
	    throw new NumberFormatException("FFT argument to small!");

	for(int i = 0; ; i++ )
	    if((powerOfTwo & (1 << i)) > 0)
		return i;
    }

    /**
     * Bitdrehereien
     */
    protected final static int reverseBits(int index, int numBits) {
	int rev;

	for(int i=rev=0; i < numBits; i++ ) {
	    rev = (rev << 1) | (index & 1);
	    index >>= 1;
	}

	return rev;
    }

    /**
     * Sehr schnelle iterative FFT-Version.
     *
     * Inspiriert von <A
     * HREF="http://www.intersrv.com/~dcross/fft.html">Don Cross's FFT
     * Seite</A>.
     *
     * Die Argumente mu diese Methode bereits durchgewrfelt erhalten!
     *
     */
    private static void FFT1DLoop(double[] realOut, double[] imagOut,
				  double angle_numerator) {
	/*
	**   Do the FFT itself...
	*/

	double delta_angle;
	double sm2;
	double sm1;
	double cm2;
	double cm1;
	double w;
	double ar0, ar1, ar2;
	double ai0, ai1, ai2;
	double tr, ti;     /* temp real, temp imaginary */
	final int numSamples = realOut.length;
	int i, j, k, n;
	int blockSize, blockEnd;

	blockEnd = 1;
	for(blockSize = 2; blockSize <= numSamples; blockSize <<= 1 ) {
	    delta_angle = angle_numerator / (double)blockSize;
	    sm2 = Math.sin(-2 * delta_angle);
	    sm1 = Math.sin(-delta_angle);
	    cm2 = Math.cos(-2 * delta_angle);
	    cm1 = Math.cos(-delta_angle);
	    w = 2 * cm1;
	    for ( i=0; i < numSamples; i += blockSize ) {
		ar2 = cm2;
		ar1 = cm1;

		ai2 = sm2;
		ai1 = sm1;

		for ( j=i, n=0; n < blockEnd; j++, n++ ) {
		    ar0 = w*ar1 - ar2;
		    ar2 = ar1;
		    ar1 = ar0;
			
		    ai0 = w*ai1 - ai2;
		    ai2 = ai1;
		    ai1 = ai0;

		    k = j + blockEnd;
		    tr = ar0*realOut[k] - ai0*imagOut[k];
		    ti = ar0*imagOut[k] + ai0*realOut[k];

		    realOut[k] = realOut[j] - tr;
		    imagOut[k] = imagOut[j] - ti;

		    realOut[j] += tr;
		    imagOut[j] += ti;
		}
	    }

	    blockEnd = blockSize;
	}
    }

    // #############################################################################
    // #############################################################################
    // #############################################################################

    private boolean shift = false;
    private boolean inverse = false;

    /**
     * Erzeugt eine neue FFT.
     * @param inverse: Falls <tt>true</tt>, wird die inverse DFT berechnet.
     * @param shift: Falls <tt>true</tt>, wird nach der Transformation der
     * Mittelpunkt verschoben
     */
    public FFT(boolean inverse, boolean shift) {

	this.shift = shift;
	this.inverse = inverse;
    }

    /**
     * Erzeugt eine neue FFT.
     * @param inverse: Falls <tt>true</tt>, wird die inverse DFT berechnet.
     */
    public FFT(boolean inverse) {this(inverse, false);}

    /**
     * Erzeugt eine neue Vorwaerts-FFT, die den Mittelpunkt nicht verschiebt.
     */
    public FFT() {this(false, false);}

    /**
     * Gibt an, ob diese FFT vorwaerts (<tt>false</tt>) oder rueckwaerts (<tt>true</tt>)
     * arbeitet.
     * @return <tt>true</tt>, falls diese FFT die inverse FFT berechnet.
     * @see #setInverse
     */
    public boolean isInverse() {return inverse;}

    /**
     * Legt fest, ob diese FFT vorwaerts (<tt>false</tt>) oder rueckwaerts (<tt>true</tt>)
     * arbeiten soll.
     * @param inverse gibt an, ob die Inverse FFT berechnet werden soll.
     * @see #isInverse
     */
    public void setInverse(boolean inverse) {this.inverse = inverse;}

    /**
     * Gibt an, ob diese FFT nach der Berechnung eine Mittelpunktsverschiebung 
     * durchfuehren soll. Falls ja, wird folgendes getan:
     * Nach einer Vorwaerts-FFT wird das Bild zyklisch in beiden Dimensionen um
     * die Haelfte verschoben. Dies hat zur Folge, dass das Transforierte Bild
     * den typischen FFT-Peak in der Mitte hat. Ist diese FFT eine Rueckwaerts-FFT,
     * so wird diese Mittelpunktsverschiebung vor der Berechnung durchgefuehrt.
     * Damit wird die Kompatibilitaet zwischen Vorwaerts- und Rueckwaerts-FFT wieder
     * hergestellt.
     * @return <tt>true</tt>, falls diese FFT nach/vor der Berechnung das Bild 
     * zyklisch verschiebt.
     * @see #setShifting
     */
    public boolean isShifting() {return shift;}

    /**
     * Legt fest, ob vor/nach der Berechnung eine Mittelpunktsverschiebung durchzufuehren
     * ist.
     * @param shift: gibt an, ob eine Mittelpunktsverschiebung durchgefuehrt werden soll.
     * @see #isShifting
     */
    public void setShifting(boolean shift) {this.shift = shift;}

    public Layer of(ProgressHandler ph, Layer p) throws PictureException {

	throw new PictureException("Veraltete Methode:\n" +
				   "Layer picture.FFT.of(Layer)");
    }
    
    public int getNumberOfProgressSteps(Picture p) {

	return 2 * p.getWidth();
    }

    /**
     * Berechnet die Fouriertransformierte des uebergebenen Bildes.
     * <p>
     * Diese Methode unterstuetzt den Progress-Listener-Mechanismus der Klasse
     * Layer mit <tt>2 * width</tt> Events.
     * 
     * @param p: Quellbild
     * @return <i>DFT</i>(<tt>p</tt>)
     * @exception PictureException, falls <tt>p</tt> nicht quadratisch ist, oder die
     * Breite von <tt>p</tt> keine Potenz von zwei ist.
     * @see ProgressHandler
     */
    public Picture of(Picture p) throws PictureException {

	boolean b = !inverse;
	boolean c = shift;

	int height = p.getHeight();
	int width = p.getWidth();

	if (!isPowerOfTwo(height) || !isPowerOfTwo(width))
	    throw new PictureException("Das Bild mu als Kantenlnge Zweierpotenzen haben!");

	if (p.getNumberOfLayers() == 1) p = realToComplex(p);

	if (!p.isComplex())
	    throw new PictureException("FFT ist nur auf reellen und komplexen Bilder mglich.");

	int mx = width; int mx2 = mx / 2; 
	int my = height; int my2 = my / 2; 

	int numSamples, numBits;
	int i, j;
	double angle_numerator = 2.0 * Math.PI;
	double[] realOut;
	double[] imagOut;
	double denom;
	double[][] picRe = new double[width][height];
	double[][] picIm = new double[width][height];

	if(inverse)
	    angle_numerator = -angle_numerator;

	/*
	 *    Do simultaneous data copy and bit-reversal ordering into
	 *    outputs...  */

	Complex t;

	/* 1. Schritt: Die komplexen Zahlen rueberschaufeln,
	 * ggf. dabei den Mittelpunkt verschieben.
	 * FFT ber Spalten */
	numSamples = height;
	numBits = numberOfBitsNeeded(numSamples);
	realOut = new double[numSamples];
	imagOut = new double[numSamples];
	denom = (double)numSamples;

	Layer re = p.getLayer(0); Layer im = p.getLayer(1);

	for(int x = 0; x < width; x++) {
	    p.fireProgressEvent();
	    for(i = 0; i < numSamples; i++) {
		j = reverseBits(i, numBits);
		realOut[j] = 
		    (b|!c) ? re.getValue(x, i) : re.getValue((mx2+x)%mx, (my2+i)%my);
		imagOut[j] = 
		    (b|!c) ? im.getValue(x, i) : im.getValue((mx2+x)%mx, (my2+i)%my);
	    }
	    FFT1DLoop(realOut, imagOut, angle_numerator);
	    if(inverse) {
		for(i=0; i < numSamples; i++) {
		    realOut[i] /= denom;
		    imagOut[i] /= denom;
		}
	    }
	    for(i = 0; i < height; i++) {
		picRe[x][(numSamples-i)%numSamples] = realOut[i];
		picIm[x][(numSamples-i)%numSamples] = imagOut[i];
	    }
	}


	/* 2. Schritt: FFT ueber die Zeilen.  Dann zurckschaufeln und
	 * ggf. Mittelpunkt verschieben. */
	numSamples = width;
	numBits = numberOfBitsNeeded(numSamples);
	realOut = new double[numSamples];
	imagOut = new double[numSamples];
	denom = (double)numSamples;

	double[][][] data = new double[2][width][height];

	for(int y = 0; y < height; y++) {
	    p.fireProgressEvent();
	    for(i=0; i < numSamples; i++) {
		j = reverseBits (i, numBits);
		realOut[j] = picRe[i][y];
		imagOut[j] = picIm[i][y];
	    }
	    FFT1DLoop(realOut, imagOut, angle_numerator);
	    if(inverse) {
		for(i=0; i < numSamples; i++) {
		    realOut[i] /= denom;
		    imagOut[i] /= denom;
		}
	    }
	    if(!b || !c)
		for(i = 0; i < width; i++) {
		    data[0][(numSamples - i) % numSamples][y] = realOut[i];
		    data[1][(numSamples - i) % numSamples][y] = imagOut[i];
		}
	    else
		for(i = 0; i < width; i++) {
		    data[0][(((numSamples - i) % numSamples) + mx2) % mx][(y + my2) % my] = 
			realOut[i];
		    data[1][(((numSamples - i) % numSamples) + mx2) % mx][(y + my2) % my] = 
			imagOut[i];
		}
	}

	return new Picture(p, 
				     new Layer[]{new Layer(data[0]), new Layer(data[1])});
    }    

    private Picture realToComplex(Picture p) {

	double[][] d = new double[p.getWidth()][p.getHeight()];

	Layer[] pics = new Layer[]{p.getLayer(0), new Layer(d)};

	return new Picture(p, pics) {
	    
	    public boolean isComplex() {return true;}
	};
    }
}
