/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.filter;

import javax.swing.JFrame;
import picture.gui.*;
import picture.*;
import picture.math.*;

/**
 * Diese Klasse erlaubt die Filterung von Layer-Objekten mit Hoch-, Tief- und
 * Bandpassfiltern. Zusaetzlich ist die Verwendung allgemeiner Filterfunktionen
 * sowie die Verwendung von Filterdateien moeglich. 
 */
public class PassFilter extends DefaultTransformation {

    /** Kennzeichnet einen Bandpassfilter */
    public static final int BANDPASSFILTER = 0;

    /** Kennzeichnet einen Hochpassfilter */
    public static final int HIGHPASSFILTER = 1;
    
    /** Kennzeichnet einen Tiefpassfilter */
    public static final int LOWPASSFILTER = 2;

    /** Kennzeichnet die Verwendung einer Filterdatei */
    public static final int FILE = 3;
    
    /** Kennzeichnet die Verwendung einer Filterfunktion */
    public static final int TERM = 4;

    private static final FFT fft = new FFT(false, false);
    private static final FFT fftI = new FFT(true, false);
    private static final FundamentalOperation mal = 
	new FundamentalOperation(FundamentalOperation.COMPLEX_PRODUCT);

    private int mode;

    private double min = 1.0 / 3;
    private double max = 2.0 / 3;
    private boolean graduell = false;

    private Term f;

    private Picture pic;

    /**
     * Erzeugt einen neuen Filter. Als Filtergrenzen werden die Standardwerte 0.33 und 0.66
     * genommen.
     * @param mode: eins von BANDPASSFILTER, HIGHPASSFILTER, LOWASSFILTER.
     * @exception IllegalArgumentException, falls mode nicht einen der oben genannten Werte hat.
     */
    public PassFilter(int mode) throws IllegalArgumentException {setMode(mode);}

    /**
     * Erzeugt einen neuen Bandpassfilter mit den angegebenen Grenzen
     * Die Werte min und max sollten zwischen 0 und 1 liegen.
     * @param min Untergrenze
     * @param max Obergrenze
     */
    public PassFilter(double min, double max) {mode = BANDPASSFILTER; setBounds(min, max);}

    /**     
     * Erzeugt einen neuen Passfilter mit der durch das angegebene PictureModel
     * definierte Filterfunktion.
     * Es wird angenommen, das der Ursprung des Koordinatensystems in der Mitte des
     * Bildes liegt.
     * 
     * @param picture: implizit gegebene Filterfunktion
     */
    public PassFilter(Picture picture) {setPicture(picture);}

    /**
     * Erzeugt einen neuen Passfilter mit der angegebenen Filterfunktion.
     * Die Funktion wird ausgewertet auf dem Intervall [0, 1].
     */
    public PassFilter(Term t) {setFunction(t);}

    /**
     * Setzt Unter- und Obergrenze, wie sie fuer einen Bandpassfilter verwendet 
     * werden. Im Falle eines Tiefpassfilters ist die Obergrenze, im Falle eines
     * Hochpassfilters die Untergrenze entscheidend.
     */
    public void setBounds(double min, double max) {

	if (min > max) {
	    double m = min;
	    min = max;
	    max = m;
	}

	this.min = min;
	this.max = max;
    }

    /**
     * Gibt Ober- und Untergrenze zurueck, die fuer Band- Tief- und Hochpassfilter
     * verwendet werden. 
     */
    public double[] getBounds() {

	return new double[]{min, max};
    }

    /**
     * Legt den Modus fest.
     * @param mode: eins von BANDPASSFILTER, LOWPASSFILTER, HIGHPASSFILTER.
     * @exception IllegalArgumentException, falls mode keinen der oben genannten Werte hat.
     */
    public void setMode(int mode) throws IllegalArgumentException {

	if (mode != BANDPASSFILTER && mode != HIGHPASSFILTER && mode != LOWPASSFILTER) 
	    throw new IllegalArgumentException("Ungltige Filterart");

	this.mode = mode;
    }

    /**
     * Gibt den aktuellen Modus zurueck. Dieser kann sein:
     * BANDPASSFILTER, LOWAPASSFILTER, HIGHPASSFILTER, FILE, TERM
     */
    public int getMode() {return mode;}

    /**
     * Legt das PictureModel fest, das zur impliziten Definition der Filterfunktion
     * herangezogen werden soll. 
     */
    public void setPicture(Picture picture) {mode = FILE; pic = picture;}

    /**
     * Falls der aktuelle Modus FILE ist, gibt diese Methode das PictureModel zurueck,
     * mit dem gefiltert wird. 
     */
    public Picture getPicture() {return pic;}

    /** Legt die Filterfunktion fest. */
    public void setFunction(Term t) {mode = TERM; this.f = t;}

    /**
     * falls der aktuelle Modus TERM ist, gibt diese Methode die Filterfunktion zurueck,
     * die angegeben wurde. 
     */
    public Term getFunction() {return f;}

    /**
     * Legt fest, ob die Filterung im Falle von Band-, Tief- oder Hochpassfiltern
     * graduell sein soll oder nicht.
     */
    public void setGraduell(boolean b) {graduell = b;}

    /**
     * gibt an, ob dieser Filter graduell ist. Ob ein Filter graduell ist oder nicht, hat
     * nur Einfluss bei Hoch- Tief- und Bandpassfiltern
     */
    public boolean isGraduell() {return graduell;}

    /**
     * erzeugt ein PGM-Bild der Groesse <tt>width x height</tt>, das den aktuell
     * gesetzten Optionen dieses Objekts entspricht.
     * Dabei wird der Ursprung in die Mitte des Bildes gelegt.
     */
    public FilterPicture createFilterPicture(int width, int height) throws PictureException {

	return (FilterPicture) createFilter(width, height).getViewableInstance();
    }

    private FilterPicture createFilter(int width, int height) throws PictureException {

	FilterPicture p = null;

	switch (mode) {
	case BANDPASSFILTER:
	case HIGHPASSFILTER:
	case LOWPASSFILTER:
	    p = new FilterPicture(mode, min, max, width, height, graduell);
	    break;
	case TERM:
	    p = new FilterPicture(f, width, height);
	    break;
	case FILE:
	    p = (FilterPicture) new FilterPicture(pic).getScaledInstance(width, height);
	    break;
	}

	return p;
    }

    public int getNumberOfProgressSteps(Picture p) {

	return 4 * p.getWidth();
    }

    /**
     * Jagt das Bild <tt>p</tt> durch diesen Filter und gibt das Ergebnis zurueck.
     * Die Filterart ist festgelegt durch die letzte Angabe von Daten. Bsp: Wurde dieser
     * Filter mit dem Konstruktor 
     * <pre>
     *  PassFilter(int, double, double)
     * </pre>
     * erzeugt und anschliessend die Methode
     * <pre>
     *  setFunction(Term)
     * </pre>
     * ausgefuehrt, so wird zur Filterung der in der Methode <tt>setFunction</tt> angegebene
     * Term verwendet. Ein Sonderfall ist die Methode <tt>setBounds</tt>. Diese setzt
     * die Filterart stets auf Bandpassfilter. Wuenscht man jedoch eine andere Filterart,
     * so steht dafuer die Methode <tt>setMode</tt> zur Verfuegung, mit der Tief- und
     * Hochpassfilter eingestellt werden koennen.
     * <p>
     * Die Filterung ist wie folgt definiert: Es sei <tt>f</tt> das Bild, das nach der
     * oben beschriebenen Art erzeugt wurde und <tt>p</tt> das zu filternde Bild. Dann
     * liefert diese Funktion das Bild
     * <center>
     * <i>FFT</i><sup>-1</sup>(<i>FFT</i>(<tt>p</tt>) * <tt>f</tt>),
     * </center>
     * wobei mit * die pixelweise Multiplikation gemeint ist.
     * <p>
     * Diese Methode unterstuetzt den ProgressHandler-Mechanismus der Klasse Layer mit
     * 4 * <tt>p.getWidth()</tt> Events.
     * 
     * @param p Quellbild
     * @return Ergebnisbild.
     * @exception PictureException z.B. wenn die Filterfunktion ungueltig ist.
     */
    public Picture of(Picture p) throws PictureException {

	ProgressHandler ph = p.getProgressHandler();
	if (p == alteEingabe) 
	    p = alteFFT;
	else {
	    alteEingabe = p;
	    p = fft.of(p);
	    alteFFT = p;
	}
	Picture q = createFilter(p.getWidth(), p.getHeight());
	p = mal.of(p, q);
	p.setProgressHandler(ph);
	return fftI.of(p);
    }

    public Layer of(ProgressHandler ph, Layer p) throws PictureException {
	
	throw new PictureException("Passfilter knnen nur auf Ebenenbildern angewandt werden.");
    }

    // Als Cache
    private transient Picture alteEingabe;
    private transient Picture alteFFT;

    /**
     * Liefert einen Optionendialog zurck.
     *
     * @param owner Parentframe
     * @param pic zu berbeitendes Bild
     * @return OptionenDialog
     */
    public DialogPane filterOptionPane(JFrame owner, Picture pic) 
	throws PictureException {

	return new FilterPane(owner, pic, this);
    }    
}
