/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.gui;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import picture.*;

/**
 * Oberklasse von ScalePane, FilterPane, SavePane  usw.
 */
public abstract class DialogPane extends JDialog {

    public static final int OK = 1;
    public static final int ABBORD = 2;
    public static final int WINDOW_CLOSING = 3;

    /**
     * Mit dieser Klasse wird die Vorschauaktualisierung gesteuert. Mit einer 
     * Methode add(Transformation) fuegt einen neuen Auftrag an. 
     * Die Klasse hat das Recht, mehrere Auftraege zu einem zusammenzufassen.
     * Bei Schliessen des Fensters kann mit kill() der Thread gestoppt werden.     
     */
    protected final class PreviewUpdater extends Thread {

	private boolean stirb = false;
	private boolean blockiert = false;
	private Vector auftraege = new Vector();
	private long auftragszeit = 0;

	public PreviewUpdater() {setPriority(NORM_PRIORITY - 1); start();}

	public synchronized void add(Transformation t) {

	    if (t == null) return;

	    synchronized (auftraege) {
		auftraege.addElement(t);
		auftragszeit = System.currentTimeMillis() + 100;
	    }

	    blockiert = false;
	    notify();
	}

	public synchronized void kill() {stirb = true; blockiert = false; notify();}

	public void run() {
	    
	    while (!stirb) 
		try {
		    abarbeiten();
		} catch (InterruptedException e) {
		} catch (PictureException e) {
		    if (changeCursor) 
			this1.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		}	    
	}

	private void abarbeiten() throws InterruptedException, PictureException {

	    Transformation t;
	    
	    // Ein Auftrag ist eingetroffen. Warte, bis 50 Millisekunden
	    // lang kein neuer Auftrag hinzukam
	    while (auftraege.size() == 0 || System.currentTimeMillis() < auftragszeit)
		if (stirb) return;
		else yield();
		
	    // nimm den neusten Auftrag und vergiss alle anderen
	    synchronized (auftraege) {
		t = (Transformation) auftraege.lastElement();
		auftraege.removeAllElements();
	    }

	    // Dieser Auftrag wird abgearbeitet.
	    if (changeCursor) this1.setCursor(new Cursor(Cursor.WAIT_CURSOR));
	    jpreview.setPictureModel(t.of(sourcePreview).getPreview());
	    jpreview.repaint();
	    if (changeCursor) this1.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));

	    // warte auf einen Auftrag oder das Ende
	    blockiert = true;		
	    sleep(50);
	    synchronized(this) {
		while (blockiert)
		    wait();
	    }
	}
    } 

    /**
     * Beinhaltet die Eventlistener
     */
    private class EventListener extends WindowAdapter implements ActionListener {

	public void windowClosing(WindowEvent e) {
	
	    updater.kill();
	}

	public void actionPerformed(ActionEvent e) {

	    String k = e.getActionCommand();

	    if (k.equals("OK")) {
		closingCode = OK;
		updater.kill();
		dispose();
	    } else if (k.equals("Abbrechen")) {
		closingCode = ABBORD;
		updater.kill();
		dispose();
	    } else if (k.equals("Speichern")) {
		DefaultTransformation dt = (DefaultTransformation)getTransformation();
		GUIHelp.saveFilter(this1, dt, new File(dt.getPreferredFileName()));
	    }
	}
    }

    /**
     * Gibt an, wie der Dialog beendet wurde.     
     */
    protected int closingCode = WINDOW_CLOSING;

    /**
     * true, um zu erreichen, dass waehrend der Berechnung der Vorschau 
     * auf WAIT_CURSOR (Sanduhr) geschaltet wird.
     */
    public boolean changeCursor = false;

    protected final PreviewUpdater updater = new PreviewUpdater();

    /** Quellbild */
    protected final Picture source;

    /** Vorschaubild von source, wird als Quelle fuer die Vorschau verwendet */
    private Picture sourcePreview; 

    /** Wird als Ziel fuer das Vorschaubild verwendet */
    private JPicture jpreview;

    private Container rahmen = new JPanel(new BorderLayout(10, 10));

    private JDialog this1 = this;

    /**
     * Erzeugt ein neues Dialogpane, das nicht vorbereitet auf einen
     * Filterdialog.
     * (aus Kompatibilitaetsgruenden)
     */
    public DialogPane(Frame owner, String title) {

	super(owner, title, true);
	rahmen = super.getContentPane();
	addWindowListener(new EventListener());
	updater.kill();
	source = null;
    }

    /**
     * Erzeugt ein neues DialogPane. Oben werden Vorschaubilder angebracht,
     * unten Bedienelemente OK und Abbrechen. Diese werden mit Leben
     * gefuellt. Eingabelemente sind in this.getContentPane().add("Center", ...)
     * einzufuegen.
     */
    public DialogPane(Frame owner, String title, Picture source) 
	throws PictureException {

	super(owner, title, true);

	if (source == null)
	    throw new PictureException("Ungltiges Bild.");
	this.source = source;
	sourcePreview = source.getPreview();

	JPicture original = new JPicture(sourcePreview);
	jpreview = new JPicture(sourcePreview);

	Container c = super.getContentPane();

	JPanel oben = new JPanel(new BorderLayout(10, 10));
	JPanel mitte = new JPanel(new BorderLayout());
	JPanel unten = new JPanel(new GridLayout(1, 5, 10, 10));
	
	JPanel bild1 = new JPanel(new BorderLayout());
	JPanel bild2 = new JPanel(new BorderLayout());

	JButton ok = new JButton("OK");
	JButton abbruch = new JButton("Abbrechen");
	JButton speichern = new JButton("Speichern");

	c.setLayout(new BorderLayout());

	oben.setBorder(new EmptyBorder(10, 10, 10, 10));

	bild1.add("Center", original);
	bild1.add("South", new JLabel("Original", JLabel.CENTER));

	bild2.add("Center", jpreview);
	bild2.add("South", new JLabel("Vorschau", JLabel.CENTER));

	oben.add("West", bild1);
	oben.add("East", bild2);

	mitte.setBorder(new TitledBorder(BorderFactory.createEtchedBorder(), 
					 "Einstellungen"));
	mitte.add("Center", rahmen);

	unten.setBorder(new EmptyBorder(10, 10, 10, 10));
	unten.add(ok);
	unten.add(abbruch);
	unten.add(speichern);
	unten.add(new JPanel());

	c.add("North", oben);
	c.add("Center", mitte);	
	c.add("South", unten);

	/*
	 *  Verkabelung
	 */
	
	EventListener ev = new EventListener();

	ok.setActionCommand("OK"); ok.addActionListener(ev);
	abbruch.setActionCommand("Abbrechen"); abbruch.addActionListener(ev);
	speichern.setActionCommand("Speichern"); speichern.addActionListener(ev);
	addWindowListener(ev);
    }

    /**
     * Layoutet dieses Fenster und setzt es in die Mitte des Bildschrims.
     * Sollte von den Unterklassen am Ende des Konstruktors aufgerufen werden.
     */
    public void prepareWindow() {

	pack();
	Toolkit tk = Toolkit.getDefaultToolkit();
	Dimension screen = tk.getScreenSize();
	Dimension size = getSize();
	setLocation((screen.width - size.width) / 2, (screen.height - size.height) / 2);
    }

    /**
     * Gibt das Quellbild zurueck
     */
    public Picture getSource() {return source;}

    /**
     * In manchen Faellen ist es sinnvoll, die Filtervorschau nicht vom Previewbild zu
     * berechnen, sondern von einem speziell aufbereiteten Bild oder vom Original. In
     * solchen Faellen kann mit dieser Methode das Quellbild festgelegt werden, von dem
     * die Transformation im Updater bestimmt werden soll. Die Anzeige des 
     * Original-Vorschau-Bildes bleibt davon unberuehrt.
     */
    public void setSourcePreview(Picture p) {

	sourcePreview = p;
    }

    /**
     * gibt das Panel zurueck, das innerhalb des mit "Einstellungen" ueberschriebenen
     * Rahmens angezeigt wird.
     */
    public Container getContentPane() {return rahmen;}

    /**
     * Gibt die Transformation zurueck, die mit diesem Fenster eingestellt wurde.
     */
    public Transformation getTransformation() {
	return null;
    }

    /**
     * Gibt das Pictureobjekt zurueck, das entsteht, wenn man die 
     * Filterung anhand der Benutzereinstellungen durchfuehrt.
     */
    public Picture getResultingPicture() throws PictureException {

	ProgressHandler p = source.getProgressHandler();
	Transformation t = getTransformation();

	if (t instanceof ProgressViewableTransformation && p != null) 
	    p.setBounds(0, ((ProgressViewableTransformation)t).getNumberOfProgressSteps(source));

	return getTransformation().of(source);
    };

    /**
     * Gibt den Code zurueck, mit dem dieses Fenster beendet wurde.
     * Dies kann sein: OK, ABBORD oder WINDOW_CLOSING, repraesentiert
     * in den Werten der entprechenden statischen Variablen dieser
     * Klasse
     */
    public int getClosingCode() {

	return closingCode;
    }
}
