/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.gui;

import java.awt.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

import picture.*;
import picture.filter.*;

public class WaveletKLTPane extends DialogPane {

    private class EigenValueArray extends JScrollPane
	implements ChangeListener, Runnable, ActionListener {

	boolean kltOrNot; private EigenValue[] array;
	JPanel comps = new JPanel(new GridLayout(0, 1));
	Thread update = new Thread(this);
	Vector queue = new Vector();
	boolean killthread = false;
	boolean blocked = true;
	long lastCall = System.currentTimeMillis();
	double [] v;

	public EigenValueArray(int number, boolean klt) {

	    this(number, new JComponent[number], klt);
	}

	public EigenValueArray(int number, JComponent[] charts, boolean klt) {

	    super(VERTICAL_SCROLLBAR_ALWAYS, HORIZONTAL_SCROLLBAR_NEVER);
	    update.start();
	    setViewportView(comps);
	    setBorder(new BevelBorder(BevelBorder.LOWERED));
	    this.kltOrNot = klt;
	    array = new EigenValue[number];

	    array[number - 1] = (charts[number - 1] == null) ?
		new EigenValue(1, 2) : new EigenValue(1, 2, charts[number - 1], false);

	    for (int i = number - 2; i >= 0; i--) 
		array[i] = (charts[i] == null) ? 
		    new EigenValue(1, array[i + 1]) : 
		new EigenValue(1, array[i + 1], charts[i], false);

	    for (int i = 0; i < array.length; i++) {
		comps.add(array[i]);
		array[i].addChangeListener(this);
	    }
	}

	public Dimension getPreferredSize() {

	    array[0].validate();
	    Dimension d = array[0].getPreferredSize();
	    d = new Dimension(d.width + getVerticalScrollBar().getPreferredSize().width,
			      3 * d.height);
	    if (d.width < 100) d.width = 250;
	    if (d.height < 100) d.height = 100;
	    return d;
	}

	public EigenValue getEigenValue(int i) {return array[i];}

	public synchronized void stateChanged(ChangeEvent e) {

	    v = new double[array.length];
	    for (int i = 0; i < v.length; i++)
		v[i] = array[i].getValue();
	    lastCall = System.currentTimeMillis();
	    blocked = false;
	    notify();
	}

	public synchronized void kill() {killthread = true; blocked = false; notify();}

	public void run() {

	    try {
		synchronized(this) {wait();}
	    } catch(InterruptedException e) {
	    }

	    while (!killthread)
		try {
		    abarbeiten();
		} catch(InterruptedException ex) {
		} catch(PictureException ex) {
		    GUIHelp.error(ex);
		    blocked = true; v = null; 
		}
	}

	private void abarbeiten() throws PictureException, InterruptedException {

	    Picture[] p; 

	    while (v == null || System.currentTimeMillis() - lastCall < 200) 
		if (killthread) return;
		else Thread.yield();

	    p = new Picture[3];
	    if (kltOrNot) {
		for (int i = 0; i < 3; i++)
		    p[i] = klt[i].kltToStandard(kltSource[i], v);
		kltPreview.setPictureModel(getSynthesis(p).getPreview());
		kltPreview.repaint();
	    } else {
		for (int i = 0; i < 3; i++)
		    p[i] = wavSource[i].getWeightedInstance(v);
		rgbPreview.setPictureModel(getSynthesis(p).getPreview());
		rgbPreview.repaint();
	    }

	    blocked = true; v = null; Thread.sleep(50);
	    
	    synchronized(this) {
		while (blocked)
		    wait();
	    }
	}

	public void actionPerformed(ActionEvent e) {

	    String k = e.getActionCommand();
	    if (k.equals("OK")) {
		closingCode = OK;
		dispose();
	    } else if (k.equals("Abbrechen")) {
		closingCode = ABBORD;
		dispose();
	    } 
	}
    }

    private class Chart extends JComponent {

	private double[] vals;

	public Chart(double[] vals) {

	    this.vals = vals; setDoubleBuffered(true);
	    setBorder(BorderFactory.createBevelBorder(BevelBorder.LOWERED));
	}

	public Dimension getPreferredSize() {return new Dimension(100, 60);}

	public void paintComponent(Graphics g) {

	    Dimension d = getSize();

	    g.setColor(Color.white); g.fillRect(0, 0, d.width, d.height);
	    
	    double max = 0;
	    for (int i = 0; i < vals.length; i++) 
		if (Math.abs(vals[i]) > max) max = Math.abs(vals[i]);

	    double dy = (d.height - 10) / max;
	    double dx = (d.width - 10) / vals.length;
	    
	    g.setColor(Color.red);
	    for (int i = 0; i < vals.length; i++)
		g.fillRect(6 + (int) (i * dx), d.height - 5 - (int) (Math.abs(vals[i]) * dy),
			   (int) dx - 5, 1 + (int) (Math.abs(vals[i]) * dy));		
	}
    }

    private Picture source;
    private Picture[] wavSource = new Picture[3];
    private Picture[] kltSource = new Picture[3];
    private LayerKLT[] klt = new LayerKLT[3];
    private Wavelet haar = Wavelet.getHaarWavelet(0);
    private JPicture kltPreview;
    private JPicture rgbPreview;
    private JComboBox comboBox = new JComboBox();

    private EigenValueArray kltEigenVals;
    private EigenValueArray rgbEigenVals;

    public WaveletKLTPane(Frame owner, Picture p) throws PictureException {

	super(owner, "Karhunen-Loeve-Transformation auf Waveletdetails");

	if (p.getNumberOfLayers() != 1)
	    throw new PictureException("Das Bild darf nur eine Ebene haben");

	addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {
		rgbEigenVals.kill();
		kltEigenVals.kill();
	    }	       
	});

	source = p;
	wavSource = getAnalysis(p);

	for (int i = 0; i < 3; i++) {
	    klt[i] = new LayerKLT(wavSource[i]);
	    kltSource[i] = klt[i].standardToKlt(wavSource[i]);
	}

	int depth = wavSource[0].getNumberOfLayers();
	JComponent[] labels = new JComponent[depth];
	for (int i = 0; i < depth; i++) 
	    labels[i] = new JLabel(i + ". EW:");

	rgbEigenVals = new EigenValueArray(depth, labels, false);

	for (int i = 0; i < depth; i++)
	    labels[i] = new JLabel(i + ". EW:");	    

	kltEigenVals = new EigenValueArray(depth, labels, true);

	Picture prev = p.getPreview();
	kltPreview = new JPicture(prev); rgbPreview = new JPicture(prev);

	Container c = getContentPane();	c.setLayout(new BorderLayout());
	
	JPanel kltPane = new JPanel(new BorderLayout(10, 10));
	JPanel rgbPane = new JPanel(new BorderLayout(10, 10));
	JPanel origPane = new JPanel(new BorderLayout(10, 10));
	JPanel buttons = new JPanel(new GridLayout(1, 5, 10, 10));

	JButton ok = new JButton("OK");
	JButton abbrechen = new JButton("Abbrechen");

	comboBox.addItem("Std-Basis");
	comboBox.addItem("KLT-Basis");
	comboBox.setSelectedItem("KLT-Basis");

	c.add("South", buttons);
	c.add("East", kltPane);
	c.add("West", rgbPane);
	c.add("North", origPane);

	buttons.setBorder(new EmptyBorder(10, 10, 10, 10));
	buttons.add(ok);
	buttons.add(abbrechen);
	buttons.add(new JPanel());
	buttons.add(new JPanel());

	JPanel optionPane = new JPanel(new GridLayout(0, 1));
	optionPane.add(new JLabel("Zur Hervorhebung der unterschiedlichen Eigenschaften sind"));
	optionPane.add(new JLabel("Manipulationen an Standard- und KLT-Basis " + 
				  "nebeneinander mglich."));
	JPanel letzteZeile = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
	letzteZeile.add(new JLabel("Die Schaltflche OK bezieht sich auf die Darstellung zur "));
	letzteZeile.add(comboBox);

	optionPane.add(letzteZeile);

	origPane.setBorder(new EmptyBorder(10, 10, 10, 10));
	origPane.add("West", new JPicture(prev));
	origPane.add("Center", optionPane);

	JPanel rgbPaneTop = new JPanel(new BorderLayout());
	rgbPaneTop.add("East", rgbPreview);

	rgbPane.setBorder(new TitledBorder(new EtchedBorder(), "Standardbasis"));
	rgbPane.add("North", rgbPaneTop);
	rgbPane.add("Center", rgbEigenVals);
	
	JPanel kltPaneTop = new JPanel(new BorderLayout());
	kltPaneTop.add("East", kltPreview);
	kltPaneTop.add("West", new Chart(klt[0].getEigenValues()));

	kltPane.setBorder(new TitledBorder(new EtchedBorder(), "optimale KLT-Basis"));
	kltPane.add("North", kltPaneTop);
	kltPane.add("Center", kltEigenVals);

	ok.addActionListener(kltEigenVals);
	abbrechen.addActionListener(kltEigenVals);
	
	prepareWindow();
    }

    public Picture getResultingPicture() throws PictureException {

	if (comboBox.getSelectedItem().equals("KLT-Basis")) 
	    return getResultKLT();
	
	return getResultStd();
    }

    public Picture getResultKLT() throws PictureException {

	double[] v = new double[wavSource[0].getNumberOfLayers()];

	for (int i = 0; i < v.length; i++)
	    v[i] = kltEigenVals.getEigenValue(i).getValue();

	Picture[] p = new Picture[3];
	for (int i = 0; i < 3; i++)
	    p[i] = klt[i].kltToStandard(kltSource[i], v);

	return getSynthesis(p);
    }

    public Picture getResultStd() throws PictureException {

	double[] v = new double[wavSource[0].getNumberOfLayers()];

	for (int i = 0; i < v.length; i++)
	    v[i] = rgbEigenVals.getEigenValue(i).getValue();

	Picture[] p = new Picture[3];
	for (int i = 0; i < 3; i++)
	    p[i] = wavSource[i].getWeightedInstance(v);

	return getSynthesis(p);
    }

    /**
     * Berechnet die redundante Waveletzerlegung und sortiert diese so um, 
     * dass je ein Picture genau die S/D, D/S und D/D-Anteile enthaelt.
     */
    private Picture[] getAnalysis(Picture p) throws PictureException {

	Picture wavSourceAll = haar.getRedundantAnalysis(p);
	Picture [] erg = new Picture[3];

	int length = 1 + (wavSourceAll.getNumberOfLayers() - 1) / 3;
	Layer[][] details = new Layer[3][length];
	for (int i = 0; i < 3; i++) {
	    details[i][length - 1] = wavSourceAll.getLayer(0);
	    for (int j = 0; j < length - 1; j++)
		details[i][j] = wavSourceAll.getLayer(j * 3 + i + 1);
	    erg[i] = new Picture(p, details[i]);
	}	

	return erg;
    }

    /** Inverse Methode zu getAnalysis(p)*/
    private Picture getSynthesis(Picture[] p) throws PictureException {

	int length = 3 * (p[0].getNumberOfLayers() - 1) + 1;
	Layer[] l = new Layer[length];

	l[0] = p[0].getLayer(p[0].getNumberOfLayers() - 1);
	for (int j = 0; j < p[0].getNumberOfLayers() - 1; j++)
	    for (int i = 0; i < 3; i++)
		l[3 * j + i + 1] = p[i].getLayer(j);

	return haar.getRedundantSynthesis(new Picture(p[0], l));
    }
}
