/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.io;

import java.util.*;
import java.io.*;

/**
 * Filtering input stream which transforms its data with the inverse
 * Burrows-Wheeler Transformation.  For more information see <a
 * href="http://gatekeeper.dec.com/pub/DEC/SRC/research-reports/abstracts/src-rr-124.html">M.
 * Burrows and D. J. Wheeler: A Block-sorting Lossless Data
 * Compression Algorithm</a>.
 *
 * This implementation is based on ideas of <a
 * href="http://www.dogma.net/markn/articles/bwt/bwt.htm">c source
 * code from Mark Nelson</a>.
 *
 * @version 0.1
 * @author Stefan Rhrich */
public class BWTInputStream extends FilterInputStream {
    private int[] count = new int[256];
    private int[] runningTotal = new int[256];

    private byte[] buf = null;
    private int pos = 0;
    private boolean end = false;

    public BWTInputStream(InputStream in) {
	super(in);
    }

    /**
     * Do the real work ...
     */
    protected void bwtInverseTransform(byte[] b, int first) throws IOException {
	buf = new byte[b.length];
	pos = 0;

	/*
 To determine a character's position in the output string given
 its position in the input string, we can use the knowledge about
 the fact that the output string is sorted.  Each character 'c' will
 show up in the output stream in in position i, where i is the sum
 total of all characters in the input buffer that precede c in the
 alphabet, plus the count of all occurences of 'c' previously in the
 input stream.

 The first part of this code calculates the running totals for all
 the characters in the alphabet.  That satisfies the first part of the
 equation needed to determine where each 'c' will go in the output
 stream.  Remember that the character pointed to by 'last' is a special
 end-of-buffer character that is supposed to be larger than any char
 in the alphabet.
	*/
        int i;

	int[] T = new int[b.length];

        for (i = 0; i < 256; i++)
            count[i] = 0;
        for (i = 0; i < b.length; i++)
	    count[ByteUtils.unsignedByte2Int(b[i])]++;

        int sum = 0;
        for (i = 0; i < 256; i++) {
            runningTotal[i] = sum;
            sum += count[i];
	    /*	    if(count[i] > 0)
		    System.err.println((char)i + ": " + runningTotal[i]);*/
            count[i] = 0;
        }
	/*
 Now that the runningTotal[] array is filled in, I have half the
 information needed to position each 'c' in the input buffer.  The
 next piece of information is simply the number of characters 'c'
 that appear before this 'c' in the input stream.  I keep track of
 that informatin in the counts[] array as I go.  By adding those
 two number together, I get the destination of each character in
 the input buffer, and that allows me to fill in all the positions
 in the transformation vector, T[].
	*/
	int index;
        for (i = 0; i < b.length; i++) {
	    index = ByteUtils.unsignedByte2Int(b[i]);
            T[count[index] + runningTotal[index]] = i;
            count[index]++;
        }

	/*	for(int k = 0; k < b.length; k++)
		System.err.println(k + ": " + T[k]);*/
	/*
 Once the transformation vector is in place, writing the
 output is just a matter of following the indices.  Note
 that I don't have to watch for the end-of-buffer character
 here.  Since I am only outputting N characters, and I know
 it will show up at position N+1, I know it won't pop up
	*/
        int j;
        i = first;
	//	System.err.println(i + " " + b.length);
        for (j = 0; j < b.length; j++) {
	    //	    System.err.println(i);
            i = T[i];
            buf[j] = b[i];
        }
    }

    /**
     * Reads and transforms a byte from the underlying output
     * stream. Therefore it must read the whole array which was
     * transformed.
     *
     * @return tranformed byte or -1 for end of file
     * @exception IOException
     */
    public int read() throws IOException {
	//	System.err.println(".");
	if(end)
	    return -1;

	if(buf != null && pos != buf.length)
	    return ByteUtils.unsignedByte2Int(buf[pos++]);
	else {
	    // Length
	    int t = in.read();
	    if(t == -1) {
		end = true;
		return -1;
	    }
	    int len = (t&0xff) << 24;
	    t = in.read();
	    if(t == -1)
		throw new IOException("Invalid BWT data (reading length 1).");
	    len |= (t&0xff) << 16;
	    t = in.read();
	    if(t == -1)
		throw new IOException("Invalid BWT data (reading length 2).");
	    len |= (t&0xff) << 8;
	    t = in.read();
	    if(t == -1)
		throw new IOException("Invalid BWT data (reading length 3).");
	    len |= (t&0xff);

	    byte[] b = new byte[len];

	    //	    System.err.println("BWT: " + len);

	    int realsize = in.read(b);
	    
	    if(realsize != len)
		throw new IOException("Invalid BWT data: " + realsize + " instead of " + len + ".");

	    // first
	    len = 0;
	    t = in.read();
	    if(t == -1)
		throw new IOException("Invalid BWT data (reading index 0).");
	    len |= (t&0xff) << 24;
	    t = in.read();
	    if(t == -1)
		throw new IOException("Invalid BWT data (reading index 1).");
	    len |= (t&0xff) << 16;
	    t = in.read();
	    if(t == -1)
		throw new IOException("Invalid BWT data (reading index 2).");
	    len |= (t&0xff) << 8;
	    t = in.read();
	    if(t == -1)
		throw new IOException("Invalid BWT data (reading index 3).");
	    len |= (t&0xff);

	    bwtInverseTransform(b, len);

	    return ByteUtils.unsignedByte2Int(buf[pos++]);
	}
    }

    /**
     * Reads and transforms a byte array from the underlying output
     * stream. Therefore it must read the whole array which was
     * transformed.
     *
     * @param b where to write the bytes to
     * @param off starting offset
     * @param len how many bytes to read
     * @return number of bytes bytes actually read
     * @exception IOException */
    public int read(byte[] b, int off, int len) throws IOException {
	int t, i;
	for(i = 0; i < len; i++) {
	    if((t = read()) == -1)
		break;
	    b[off+i] = (byte)t;
	}
	return i;
    }

    /**
     * @return 0
     */
    public int available() throws IOException {
	return 0;
    }

    /**
     * Closes the underlying input stream and deletes own buffer.
     */
    public void close() throws IOException {
	buf = null;
	in.close();
    }

    /**
     * Do nothing.
     */
    public void mark(int readlimit) {
    }

    /**
     * Do nothing.
     */
    public void reset() throws IOException {
    }

    /**
     * @return false
     */
    public boolean markSupported() {
	return false;
    }

    /**
     * Skips bytes.
     *
     * @param n how many bytes to skip.
     * @return number of bytes actually skipped
     * @exception IOException */
    public long skip(long n) throws IOException {
	long i;
	int t;
	for(i = 0; i < n; i++)
	    if((t = read()) == -1)
		break;
	return i;
    }

    public static void main(String[] args) throws IOException {
	BWTInputStream bwt = new BWTInputStream(System.in);

	/*
	bwt.bwtInverseTransform("bbcaa".getBytes(), 0);
	System.out.println();
	bwt.bwtInverseTransform("drbauarak".getBytes(), 3);
	System.out.println();
	*/

	int t;
	while((t = bwt.read()) != -1)
	    System.out.write(t);

	System.out.flush();
	System.out.close();
    }
}
