/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.math;
import picture.Complex;
import java.util.*;
import java.io.*;

/**
 * Diese Klasse repraesentiert einen Term, also einen mathematischen Formelausdruck. Ein Term 
 * ist eine Verknuepfung aus Termen mit den Grundrechenarten oder eine Funktion von einem Term 
 * oder eine Klammerung von einem Term. Schlie&szlig; sind Terme Zusammensetzungen aus 
 * Variablen und Konstanten.
 * <p>
 * Die Klasse erlaubt die Verkn&uuml;pfung von Termen zu neuen Termen sowie die Berechnung von 
 * Funktionswerten f&uuml;r bestimmte Variablenbelegungen als Complex.
 * 
 * @version 1.0
 * @author Manuel Kauers (c) 2000
 */
// Der Code ist eine abgespeckte Version der Term-Klasse meines Computeralgebrapakets.
// (siehe http://www.uni-karlsruhe.de/~uinf/analysis/)
public class Term implements Serializable {

    // #############################################################################
    // ###  statische Variablen  ###################################################
    // #############################################################################

    private static final int M = 0;
    private static final int S = 1;
    private static final int P = 2;
    private static final int F = 3;
    private static final int V = 4;
    private static final int E = 5;
    private static final int K = 6;
    private static final int Z = 7;
    private static final int B = 8;
    private static final int Vec = 9;
    private static final int Vek = 9;

    private static final Complex eins = new Complex(1, 0);
    private static final Complex zwei = new Complex(2, 0);
    private static final Complex nul = new Complex(0, 0);    

    // #############################################################################
    // ###  Instanzvariablen  ######################################################
    // #############################################################################

    /** 
     * Waehrend einer Berechnung kann die folgende Variable auf true gesetzt werden, um zu 
     * erreichen, dass oefter als noetig vereinfacht wird.
     */    
    private static boolean vereinfachungssperre = false;

    /** enthaelt das nonterminale Symbol und damit die Bedeutung dieses Objekts */
    private int status = -1;

    /**
     * Enthaelt die Summanden im Fall status == S bzw die Faktoren des Zaehlers im Fall 
     * status == P bzw das Argument im Fall status == F bzw Basis (param(0)) und Exponent 
     * (param(1)) im Fall status == E
     */
    private Vector param = new Vector();

    /** Enthaelt die Faktoren des Nenners im Fall status == P */
    private Vector nenner = new Vector();

    /** Enthaelt den Vorfaktor */
    private Complex faktor = new Complex(1, 0);

    /** 
     * Enthaelt die Bezeichnung dieser Variable (falls status == V)
     * bzw dieser Funktion (falls status == F)
     */
    private String name;

    // #############################################################################
    // ###  Konstruktoren / Parsen  ################################################
    // #############################################################################

    /**
     * Erzeugt ein neues Term-Objekt aus dem angegebenen String. Der String enthalte eine 
     * Repraesentation des Ausdrucks in vereinfachter Maple-Syntax. Das heisst: Erkannt 
     * werden Summen ("+"), Differenzen ("-"), Produkte ("*"), Divisionen ("/"),
     * Klammerung ("(", ")"), Exponentiation ("^") sowie Funktionen (Bezeicher +  "(" + 
     * Term + ")"), Variablen (Bezeichner) und Konstanten.<br>
     * Bezeichner beginnen mit einem Buchstaben, gefolgt von beliebig vielen weiteren Ziffern
     * und Buchstaben.<br>
     * Zahlen beginnen mit einer Ziffer, gefolgt von weiteren Ziffern, optional gefolgt on 
     * einem Komma (".") und mindestens einer weiteren Ziffer.<br>
     * Die Semantik entspricht der ueblichen. 
     * @param s der Ausdruck
     * @exception IllegalArgumentException, falls s nicht verstanden wird
     */
    public Term(String s) throws IllegalArgumentException {

	this(ohneLeerzeichen(s), M);
	simplify();
    }

    /**
     * Erzeugt einen neuen Term, der eine Konstante 1 repraesentiert
     */
    public Term() {status = K;}

    /**
     * Erzeugt einen neuen Term aus den angegebenen Daten.
     */
    private Term(int status, Complex faktor, String name) {
	this.status = status; this.faktor = faktor; this.name = name;
    }

    /**
     * In diesem privaten Konstruktor geht der Parsvorgang vor sich.
     * Kommentar zur Grammatik: vgl interface Konstanten.
     */
    private Term(String s, int status) throws IllegalArgumentException {

	if (s.length() == 0)
	    throw new IllegalArgumentException();

	int i; int j; int k; int l;
	Term t = new Term(); boolean n = true; Vector p;
	
	try {
	    switch (status) {
	    case M: 
		// n legt hier fest, ob der Vorfaktor des folgenden
		// Summanden negativ ist. 
		n = (s.charAt(0) == '-');
		if (n)
		    s = s.substring(1);
		// dass hier kein break steht, ist Absicht.
	    case S: 
		this.status = S;
		for (i = 0;;) {
		    
		    j = next(s, i, '+');
		    k = next(s, i, '-');
		    j = (j > k) ? k : j; // = min(k, j)
		    
		    if (j == Integer.MAX_VALUE) 
			t = new Term(s.substring(i), P);
		    else		    
			t = new Term(s.substring(i, j), P);

		    if (n) 
			t.faktor = (Complex) t.faktor.mul(-1);

		    param.addElement(t);

		    if (j == Integer.MAX_VALUE)
			break;
		    n = (s.charAt(j) == '-');
		    i = j + 1;
		}
		break;
	    case P: 
		this.status = P;
		// n ist wahr, wenn der naechste Teil-Term zum Nenner kommt
		n = false; 
		for (i = 0; i >= 0 && i < s.length();) {

		    j = next(s, i, '*');
		    k = next(s, i, '/');
		    l = next(s, i, '^');
		    j = (j > k) ? k : j; //entscheidend ist die kleinere Zahl

		    if (l >= j) {// P -> F
			if (j == Integer.MAX_VALUE) 
			    t = new Term(s.substring(i), F);
			else
			    t = new Term(s.substring(i, j), F);
		    } else { // P -> F ^ F
			t = new Term(E, new Complex(1, 0), "");
			t.param.addElement(new Term(s.substring(i, l), F));
			if (j == Integer.MAX_VALUE) // Exp.
			    t.param.addElement(new Term(s.substring(l + 1), F));
			else
			    t.param.addElement(new Term(s.substring(l + 1, j), F));
		    }

		    if (n) {
			if (t.status == K)
			    faktor = (Complex) faktor.div(t.faktor);
			else
			    nenner.addElement(t);
		    } else {
			if (t.status == K) 
			    faktor = (Complex) faktor.mul(t.faktor);
			else
			    param.addElement(t);
		    }
		    
		    if (j == Integer.MAX_VALUE)
			break;
		    
		    n = (s.charAt(j) == '/');
		    i = j + 1;
		}
		
		// wenn es keinen Zaehler gibt, dann ist dieser eins
		if (param.size() == 0 && nenner.size() > 0)  
		    param.addElement(new Term()); // = 1

		break;
	    case F: 
		this.status = F;
		name = "";
		if (s.charAt(0) == '(') // F -> (Vec)
		    param.addElement(new Term(s.substring(1, s.length() - 1), Vec));
		else if (s.charAt(s.length() - 1) == ')') { // F -> f(Vec)
		    for (i = 0; i < s.length() && s.charAt(i) != '('; i++);
		    name = (new Term(s.substring(0, i), V)).name;
		    if (name.equals("\""))
			throw new IllegalArgumentException();
		    param.addElement(new Term(s.substring(i + 1, s.length() - 1), Vec));
		} else if (istBuchstabe(s.charAt(0))) // F -> V
		    param.addElement(new Term(s, V));
		else if (istZiffer(s.charAt(0))) // F -> K
		    param.addElement(new Term(s, K));	
		else if (s.charAt(0) == '\"')
		    param.addElement(new Term(V, new Complex(1, 0), "\""));
		else
		    throw new IllegalArgumentException();
		break;

	    case V: 
		this.status = V;
		// Ein Bezeichner ist " oder er genuegt der folgenden Grammatik:
		// S -> BK
		// K -> K[K] | KB | KZ | lambda
		// B -> Buchstabe, Z -> Ziffer	
		// Falls was anderes als Bezeichner generiert werden soll, wird eine
		// IllegalArgumentException ausgeworfen.
		char c;	name = s; j = s.length(); 
		for (i = 0; i < j; i++) {
		    c = s.charAt(i);
		    if (c == '[') {
			if (s.charAt(j - 1) == ']') 
			    j--;
			else
			    throw new IllegalArgumentException();
		    } else if (!istBuchstabe(c) && !istZiffer(c))
			throw new IllegalArgumentException();
		}
		break;
	    case K: 
		this.status = K; 
		int punktZahl = 0;
		for (i = 0; i < s.length(); i++)
		    if (!istZiffer(s.charAt(i))) {
			if (s.charAt(i) == '.')
			    punktZahl++;
			else
			    throw new IllegalArgumentException();
		    }

		double wert = (new Double(s)).doubleValue();
		if (punktZahl <= 1)
		    faktor = new Complex(wert, 0);
		else if (punktZahl > 1)
		    throw new IllegalArgumentException();
		break;

	    case Vec:
		this.status = Vec; j = 0; i = next(s, j, ','); 
		while (i != Integer.MAX_VALUE) {
		    param.addElement(new Term(s.substring(j, i), M));
		    j = i + 1; i = next(s, j, ','); 
		}
		param.addElement(new Term(s.substring(j), M));
	    }
	} catch (ArithmeticException e) {
	    throw e;
	} catch (Exception e) {
	    throw new IllegalArgumentException();
	}
    }

    // #############################################################################
    // ###  Ausgabe- und Umwandlungsmethoden  ######################################
    // #############################################################################

    // Hier aus Platzgruenden teilweise gestrichen,  
    // da die Uebertragung Zahl -> Complex zu komplex geworden waere.

    private String vorfaktor(Complex z) {

	if (z.equals(eins))
	    return "";

	double re = z.getRealPart();
	double im = z.getImaginaryPart();

	if ((re > 0 && Math.abs(im) < Complex.epsilon) ||
	    (im > 0 && Math.abs(re) < Complex.epsilon))
	    return z + " * ";

	return "(" + z + ") * ";
    }

    /**
     * Berechnet eine String-Repraesentation dieses Terms, mit dem der Konstruktor
     * etwas anfangen kann.
     */
    public String toString() {

	boolean b = true;
	String s = "";
	String auf = "(";
	String zu = ")";
	String mal = " * ";
	String lz = " ";

	switch (status) {
	case S: {
	    for (int i = 0; i < param.size(); i++) {

		Term t = pea(i);
		if (i > 0)
		    s += " + ";
		else
		    s += (i > 0) ? lz : "";

		s += t.toString();
	    }
	    if (!faktor.equals(eins))
		s = vorfaktor(faktor) + auf + s + zu;
	}; break;
	case P: {
	    Term t; String k;
	    for (int i = 0; i < param.size(); i++) {
		t = pea(i); k = t.toString();
		if (t.status == S && (param.size() > 1 || b)) 
		    k = auf + k + zu;
		s += k + ((i < param.size() - 1) ? mal : "");
	    }

	    if (nenner.size() > 0) {

		if (param.size() > 1 && b)
		    s = auf + s + zu;

		String n = "";
		for (int i = 0; i < nenner.size(); i++) {
		    t = nea(i); k = t.toString();
		    if (t.status == S)
			k = auf + k + zu;
		    n += k + ((i < nenner.size() - 1) ? mal : "");
		}

		if (nenner.size() > 1 && b)
		    n = auf + n + zu;

		s += " / " + n;
	    }
	    s = vorfaktor(faktor) + s;
	}; break;
	case F: {
	    Term t = pea(0);
	    s = t.toString();
	    if (t.status != Vec)
		s = auf + s + zu;
	    s = vorfaktor(faktor) + name + s;
	}; break;
	case E: {
	    Term t = pea(0); s = auf + t.toString() + zu;
	    s += " ^ ";
	    t = pea(1); String s0 = auf + t.toString() + zu;
	    s += s0;
	    s = vorfaktor(faktor) + s;
	}; break;
	case V: {
	    s = vorfaktor(faktor) + name;
	}; break;
	case K: {
	    s = faktor.toString();
	}; break;
	case Vec: {
	    for (int i = 0; i < param.size() - 1; i++)
		s += pea(i).toString() + ", ";
	    s += ((Term)param.lastElement()).toString();
	    s = vorfaktor(faktor) + s;
	}
	}

	//return "[" + status + ":" + s + "]";
	return s;
    }

    // #############################################################################
    // ###  Methoden zur Termvereinfachung  ########################################
    // #############################################################################

    /**
     * true, falls this und <tt>t</tt> gleich sind im Sinne gleicher Struktur.
     * Als gleich werden erkannt: permutierte Summen und Produkte, etwa liefert
     * <tt>a + b = b + c</tt> true, w&auml;hrend 
     * <tt>sin(x)^2 + cos(x)^2 = 1</tt> false liefert.
     * @param t der Term, mit dem this verglichen werd
     */
    public boolean equals(Term t) {

	if (status != t.status || !faktor.equals(t.faktor))
	    return false;

	switch (status) {
	case F: return (name.equals(t.name) && pea(0).equals(t.pea(0)));
	case V: 
	    return (name.equals(t.name) && (param.size() == t.param.size()) &&
		    (param.size() == 0 || pea(0).equals(t.pea(0))));
	case K: return true;
	case E: return (pea(0).equals(t.pea(0)) && pea(1).equals(t.pea(1)));
	case S: {
	    if (param.size() != t.param.size())
		return false;
	    boolean [] b = new boolean[param.size()];
	    for (int i = 0; i < b.length; i++)
		b[i] = false;
	    for (int i = 0; i < b.length; i++)
		for (int j = 0; j < b.length; j++)
		    if (!b[j] && pea(j).equals(t.pea(i)))
			b[j] = true;
	    for (int i = 0; i < b.length; i++)
		b[0] &= b[i];
	    return b[0];
	}
	case P: {
	    if (param.size() != t.param.size() || nenner.size() != t.nenner.size())
		return false;
	    // Zaehler vergleichen
	    boolean [] b = new boolean[param.size()];
	    for (int i = 0; i < b.length; i++)
		b[i] = false;
	    for (int i = 0; i < b.length; i++)
		for (int j = 0; j < b.length; j++)
		    if (pea(j).equals(t.pea(i)) && !b[j])
			b[j] = true;
	    for (int i = 0; i < b.length; i++)
		b[0] &= b[i];
	    // Nenner vergleichen
	    if (nenner.size() > 0) {
		boolean [] c = new boolean[nenner.size()];
		for (int i = 0; i < c.length; i++)
		    c[i] = false;
		for (int i = 0; i < c.length; i++)
		    for (int j = 0; j < c.length; j++)
			if (!c[j] && nea(j).equals(t.nea(i)))
			    c[j] = true;
		for (int i = 0; i < c.length; i++)
		    c[0] &= c[i];
		b[0] &= c[0];
	    }
	    return b[0];
	}
	case Vec: {
	    if (param.size() != t.param.size())
		return false;
	    boolean b = true;
	    for (int i = 0; i < param.size(); i++)
		b &= pea(i).equals(t.pea(i));
	    return b;
	}
	}
	return false;
    }

    /** 
     * erste Vereinfachungsstufe
     * - Kinder vereinfachen
     * - Liften
     * - Struktur vereinfachen
     * - Zusammenfassen von Summen, Produkten und Potenzen
     * - sonstiges
     * @return true, falls was geaendert wurde.
     */
    private boolean mySimplify() {

	if (status == K || vereinfachungssperre)
	    return false;

	if (faktor.equals(nul)) { // 0 * f(x) = 0
	    liften(new Term(K, new Complex(0, 0), ""));
	    return true;
	}

	boolean b = false; Term t; Term u; int i; int j;
	
	// Kinder vereinfachen
	for (i = 0; i < param.size(); i++)
	    while (pea(i).mySimplify());
	for (i = 0; i < nenner.size(); i++)
	    while (nea(i).mySimplify());

	switch (status) {
	case S: {

	    // 0. Liften
	    if (param.size() == 1) {
		liften(pea(0));
		return true;
	    }

	    // 1. Struktur vereinfachen
	    Vector p = new Vector();
	    //String v = "";
	    for (i = 0; i < param.size(); i++) {
		t = pea(i);
		if (t.status == F && t.name.equals("")) 
		    t.liften(t.pea(0));
		if (t.status == S) 
		    for (j = 0; j < t.param.size(); j++) {
			b = true; u = t.pea(j);
			u.faktor = (Complex) u.faktor.mul(t.faktor);
			p.addElement(u);
		    }
		else {
		    p.addElement(pea(i));
		}
	    }

	    param = p;

	    // 2. Zusammenfassen
	    // Es wird eine Tabelle erzeugt, in die fuer jeden Summanden
	    // die Haeufigkeit eingetragen wird. Dazu werden bei Produkten
	    // die symbolischen Terme von ihren Vorfaktoren getrennt,
	    // die Vorfaktoren zur bisherigen Summe hinzuaddiert.
	    // Zum Schluss werden die einzelnen Summanden mit ihren
	    // Vorfaktoren (falls betragsmaessig ungleich eins) wieder
	    // zu einem Summenobjekt zusammengefasst.
	    // Konstanten werden gesondert behandelt und in k[0]
	    // akkumuliert.

	    Term [] k = new Term[param.size() + 1];
	    Complex [] anz = new Complex[k.length];
	    int frei = 1; // k[1]..k[frei] sind belegt. 
	    boolean bekannt;
	    anz[0] = new Complex(0, 0);

	    for (i = 0; i < param.size(); i++) {

		t = pea(i); u = t.copy(); u.faktor = new Complex(1, 0);
		// u ist t ohne Vorfaktor
		bekannt = false;
		if (t.status == K)
		    anz[0] = (Complex) anz[0].add(t.faktor);
		else {
		    for (j = 1; j < frei; j++) 
			if (u.equals(k[j])) {
			    anz[j] = (Complex) anz[j].add(t.faktor);
			    bekannt = true;
			    break;
			}		
		    if (!bekannt) {
			k[frei] = u;
			anz[frei++] = t.faktor;
		    }
		}
	    }

	    // Jetzt wird eingesammelt: Der neue param-Vector wird erstellt
	    p = new Vector();
	    if (!anz[0].equals(nul)) {
		t = new Term(K, anz[0], "");
		p.addElement(t);
	    }
	    for (i = 1; i < frei; i++) 
		// Was Vorfaktor 0 hat, wird nicht aufgenommen.
		if (!anz[i].equals(nul)) {
		    k[i].faktor = anz[i];
		    p.addElement(k[i]);
		}

	    if (p.size() == 0) 
		p.addElement(new Term(K, new Complex(0, 0), ""));

	    b |= (p.size() < param.size());
	    param = p;

	    // Wir pruefen nun die Existenz von Vektoren: Existieren Vektoren als
	    // Summanden, so werden diese zusammengefasst ala
	    // (a, b) + (c, d) = (a + c, b + d).
	    // Falls die Dimension der Vektoren nicht uebereinstimmt, wird eine
	    // Exception ausgespuckt.
	    Term vec = null; int d = 0;
	    for (i = 0; i < param.size(); i++) {
		u = pea(i); 
		if (u.status == F && u.name.equals(""))
		    u = u.pea(0);
		if (d == 0 && u.dim() != 0)
		    d = u.dim();
		if (0 != u.dim() && d != u.dim() && d != 0)
		    throw new ArithmeticException("Dimensionsfehler in Addition");
		if (u.status == Vec) {
		    if (!u.faktor.equals(nul)) {
			for (j = 0; j < u.param.size(); j++)
			    u.pea(j).faktor = (Complex) u.pea(j).faktor.mul(u.faktor);
			u.faktor = new Complex(1, 0);
		    }
		    if (vec == null) 
			vec = u; 
		    else {
			b = true; p = new Vector();
			for (j = 0; j < d; j++) 
			    p.addElement(vec.pea(j).plus(u.pea(j)));
			vec.param = p;
		    }
		    param.removeElementAt(i); i--;
		}
	    }
	    if (vec != null)
		param.insertElementAt(vec, 0);

	}; break;
	case P: {

	    // 0. Liften
	    if (nenner.size() == 0 && param.size() == 1) {
		liften(pea(0));		
		return true;
	    } 

	    if (param.size() == 0) {
		// this besteht nur aus einem Vorfaktor
		status = K; 
		return true;
	    }
		
	    // 1. Struktur vereinfachen
	    int paramSize = param.size();
	    int nennerSize = nenner.size();
	    Vector p = new Vector();
	    Vector n = new Vector();
		    
	    for (i = 0; i < param.size(); i++) {
		t = pea(i);
		if (t.status == F && t.name.equals("") && t.pea(0).status != S)
		    t.liften(t.pea(0));
		if (t.status == P) {
		    for (j = 0; j < t.param.size(); j++) 
			p.addElement(t.pea(j));
		    for (j = 0; j < t.nenner.size(); j++)
			n.addElement(t.nea(j));
		} else if (t.status != K)
		    p.addElement(t);
		faktor = (Complex) faktor.mul(t.faktor);
		t.faktor = new Complex(1, 0);
	    }

	    for (i = 0; i < nenner.size(); i++) {
		t = nea(i);
		if (t.status == F && t.name.equals("") && t.pea(0).status != S)
		    t.liften(t.pea(0));
		if (t.status == P) {
		    for (j = 0; j < t.param.size(); j++)
			n.addElement(t.pea(j));
		    for (j = 0; j < t.nenner.size(); j++)
			p.addElement(t.nea(j));
		} else if (t.status != K)
		    n.addElement(t);
		faktor = (Complex) faktor.div(t.faktor);
		t.faktor = new Complex(1, 0);
	    }

	    if (faktor.equals(nul)) { // 0 * f(x) = 0
		status = K;
		param = new Vector();
		nenner = new Vector();
		return true;
	    }

	    param = p; 
	    nenner = n;	    
	    // Gucken, ob da jemand durch einen Vektor teilen will.
	    for (i = 0; i < n.size(); i++)
		if (nea(i).dim() > 1)
		    throw new ArithmeticException("Division durch Vektor: " + nea(i));

	    // Veraendere sqrt(..) zu ..^(1/2)
	    for (i = 0; i < p.size(); i++) {
		t = pea(i);
		if (t.status == F && t.name.equals("sqrt")) {
		    t.status = E;
		    t.param.addElement(new Term(K, new Complex(0.5, 0), ""));
		}		
	    }
	    for (i = 0; i < n.size(); i++) {
		t = nea(i);
		if (t.status == F && t.name.equals("sqrt")) {
		    t.status = E;
		    t.param.addElement(new Term(K, new Complex(0.5, 0), ""));
		}
	    }

	    // 2. Zusammenfassen
	    // Strategie wie bei Summe, Kommentar siehe dort.
	    // Oben wurden bereits die Vorfaktoren verrechnet, so
	    // dass hier davon ausgegangen werden kann, dass alle
	    // Faktoren den Vorfaktor eins haben.
	    Term [] k = new Term[param.size() + nenner.size()];
	    Complex [] exp = new Complex[k.length];
	    int frei = 0; boolean bekannt = false;
	    for (int vorzeichen = 1; vorzeichen > -2; vorzeichen -= 2) {

		if (vorzeichen == 1) p = param; else p = nenner;

		for (i = 0; i < p.size(); i++) {
		    
		    t = (Term) p.elementAt(i);

		    bekannt = false;
		    if (t.status == E && t.pea(1).status == K) {
			u = t.pea(1);
			if (u.status == F)
			    u = u.pea(0);
			for (j = 0; j < frei; j++) {
			    if (k[j].equals(t.pea(0))) {
				exp[j] = (Complex) exp[j].add(u.faktor.mul(vorzeichen));
				bekannt = true;
				break;
			    }
			}
			if (!bekannt) {
			    k[frei] = t.pea(0);
			    u = t.pea(1);
			    exp[frei++] = (Complex) u.faktor.mul(vorzeichen);
			}
		    } else if (t.status != K) {
			for (j = 0; j < frei; j++) 
			    if (k[j].equals(t)) {
				exp[j] = (Complex) exp[j].add(vorzeichen);
				bekannt = true;
				break;
			    }
			if (!bekannt) {
			    k[frei] = t;
			    exp[frei++] = new Complex(vorzeichen, 0);
			}
		    }
		}
	    }

	    p = new Vector();
	    n = new Vector();

	    for (i = 0; i < frei; i++) 
		if (!exp[i].equals(nul)) {
		    t = new Term(E, new Complex(1, 0), "");
		    t.param.addElement(k[i]);
		    t.param.addElement(new Term(K, 
						new Complex(exp[i].getRealPart(), 0), 
						""));
		    t.mySimplify();
		    p.addElement(t);
		}

	    if (p.size() == 0) 
		p.addElement(new Term());

	    b |= (p.size() != paramSize || n.size() != nennerSize);
	    param = p;
	    nenner = n;

	    // Vektoren zaehlen
	    int anzV = 0; Term [] vek = new Term[2];
	    try {
		for (i = 0; i < param.size(); i++)
		    if (pea(i).dim() > 1) {
			vek[anzV++] = pea(i);
			param.removeElementAt(i--);
		    }
	    } catch (ArrayIndexOutOfBoundsException e) {
		// mehr Vektoren: Fehlermeldung
		throw new ArithmeticException("Mehr als zwei Vektoren im Produkt");
	    }
	    if (anzV == 1 && vek[0].status == Vec) {
		// ein Vektor: Liften, falls explizit; also:
		// x * (y, z) -> (x * y, x * z)
		b = true; p = new Vector(); param.addElement(new Term()); 
		for (i = 0; i < vek[0].param.size(); i++)
		    p.addElement(mal(vek[0].pea(i)));
		vek[0].param = p; faktor = new Complex(1, 0);
		liften(vek[0]);				 
	    } else if (anzV == 2 && vek[0].status == Vec && vek[1].status == Vec) {
		if (vek[0].dim() != vek[1].dim())
		    throw new ArithmeticException("Dimensionsfehler in Skalarprodukt");
		// zwei Vektoren: Skalarprodukt, falls explizit; also:
		// (a, b) * (c, d) -> a * c + b * d
		b = true; u = new Term(S, new Complex(1, 0), "");
		for (i = 0; i < vek[0].param.size(); i++)
		    u.param.addElement(vek[0].pea(i).mal(vek[1].pea(i)));
		param.insertElementAt(u, 0); //einbetten(u), 0);
	    }

	}; break;
	case F: {

	    t = pea(0);

	    //0. Liften
	    while (t.status == F && t.name.equals("")) 
		t.liften(t.pea(0));

	    if (name.equals("")) {
		liften(t); 
		return true;
	    } else if (t.status == K || t.status == Vec) {
		
		Complex erg = null;
		if (t.status == K) {
		    if (name.equals("sin")) erg = sin(t.faktor);
		    else if (name.equals("cos")) erg = cos(t.faktor);
		    else if (name.equals("tan")) erg = tan(t.faktor);
		    else if (name.equals("sinh")) erg = sinh(t.faktor);
		    else if (name.equals("cosh")) erg = cosh(t.faktor);
		    else if (name.equals("tanh")) erg = tanh(t.faktor);
		    else if (name.equals("exp")) erg = exp(t.faktor);
		    else if (name.equals("log")) erg = log(t.faktor);
		    else if (name.equals("arcsin")) erg = arcsin(t.faktor);
		    else if (name.equals("arccos")) erg = arccos(t.faktor);
		    else if (name.equals("arctan")) erg = arctan(t.faktor);
		    else if (name.equals("arcsinh")) erg = arcsinh(t.faktor);
		    else if (name.equals("arccosh")) erg = arccosh(t.faktor);
		    else if (name.equals("arctanh")) erg = arctanh(t.faktor);
		    else if (name.equals("sqrt")) erg = sqrt(t.faktor);
		    else if (name.equals("abs")) erg = abs(t.faktor);
		    else if (name.equals("Re")) erg = re(t.faktor);
		    else if (name.equals("Im")) erg = im(t.faktor);	    
		}
		
		// Eingeflickt: Indikatorfunktion
		if (t.status == Vec && t.dim() == 3 && t.pea(0).status == K &&
		    t.pea(1).status == K && t.pea(2).status == K)
		    erg = rect(t.pea(0).faktor, t.pea(1).faktor, t.pea(2).faktor);
		
		if (erg != null) {
		    faktor = faktor.mul(erg);
		    status = K;
		}
	    } 
	}; break;
	case E: {

	    t = pea(0); u = pea(1);
	    if (u.status == F && u.name.equals(""))
		u.liften(u.pea(0));

	    if (t.status == K && u.status == K) {
		//if (!t.faktor.istExakt() || !u.faktor.istExakt()) {
		    // unexakt ausrechnen
		if (t.faktor.equals(nul)) {
		    if (u.faktor.equals(nul))
			t.faktor = new Complex(1, 0);
		    else 
			t.faktor = new Complex(0, 0);
		} else
		    t.faktor = exp((Complex) u.faktor.mul(log(t.faktor)));
		liften(t); 
		return true;
	    }

	    Complex d = t.faktor;
	    // a ^ f(x) = a  (a = 0, 1)
	    if (t.status == K && (d.equals(eins) || d.equals(nul))) {
		liften(t);
		return true;
	    }

	    if (u.status == K) {

		d = u.faktor;
		if (d.equals(nul)) {// f(x) ^ 0 = 1
		    liften(new Term());
		    return true;
		} else if (d.equals(eins)) { // f(x) ^ 1 = f(x)
		    liften(pea(0));
		    return true;
		} else if (u.faktor.equals(new Complex(0.5, 0))) {
		    // f(x) ^ (1/2) = sqrt(f(x))
		    u = new Term(F, faktor, "sqrt");
		    u.param.addElement(t);
		    liften(u);
		    return true;
		} else if (u.faktor.equals(zwei) && t.status == F &&
			   t.name.equals("sqrt")) {
		    // sqrt(f(x))^2 = f(x)
		    liften(t.pea(0));
		    return true;
		} else if (t.status != K && !t.faktor.equals(eins)) {

		    // (konst * f(x)) ^ c = konst ^ c * f(x) ^ c
		    Term t0 = new Term(P, new Complex(1, 0), "");
		    Term t1 = new Term(E, new Complex(1, 0), "");
		    t1.param.addElement(new Term(K, t.faktor, ""));
		    t1.param.addElement(u.copy());
		    t0.param.addElement(t1);
		    t1 = new Term(E, new Complex(1, 0), "");
		    t1.param.addElement(t);
		    t1.param.addElement(u);
		    t.faktor = new Complex(1, 0);
		    t0.param.addElement(t1);
		    liften(t0);
		    return true;
		}
	    }

	    if (t.status == V && t.name.equals("e")) {
		// e^f(x) = exp(f(x))
		t = new Term(F, new Complex(1, 0), "exp");
		t.param.addElement(u);
		liften(t);
		return true;
	    }

	    if (t.status == P && u.status == K) {
		// (a * b)^c = a^c * b^c
		Vector p = new Vector(); Vector n = new Vector();
		for (i = 0; i < t.param.size(); i++) {
		    Term v = new Term(E, new Complex(1, 0), "");
		    v.param.addElement(t.pea(i));
		    v.param.addElement(u.copy());
		    p.addElement(v);
		}		    
		for (i = 0; i < t.nenner.size(); i++) {
		    Term v = new Term(E, new Complex(1, 0), "");
		    v.param.addElement(t.nea(i));
		    v.param.addElement(u.copy());
		    n.addElement(v);
		}
		t.param = p; t.nenner = n; liften(t); 
		return true;
	    }

	    if (t.status == E) {
		// (a^b)^c = a^(b*c)
		Term v = pea(1).mal(t.pea(1));
		t.param.removeElementAt(1);
		t.param.addElement(v);
		liften(t); 
		return true;
	    }

	    if (t.status == Vec) {
		if (u.status != K || !u.faktor.equals(zwei)) 
		    throw new ArithmeticException("Potenzierung eines Vektors");
		Vector p = new Vector();
		for (i = 0; i < t.dim(); i++)
		    p.addElement(t.pea(i).mal(t.pea(i)));
		t.status = S; t.param = p; liften(t); //einbetten(t)); 
		return true;
	    }

	    if (u.dim() > 1)
		throw new ArithmeticException("Dimensionsfehler in Exponent");

	    Vector v = new Vector(); v.addElement(t); v.addElement(u);
	    param = v;

	}; break;
	case V: {
	    if (faktor.equals(nul)) {
		name = ""; status = K; b = true;
	    } else if (name.toLowerCase().equals("i")) {
		name = ""; status = K; b = true;
		faktor = (Complex) faktor.mul(Complex.i);
	    } 
	    /*
	    u = Variablenliste.holeTerm(name);
	    if (u != null) {
		liften(u); b = true;
	    }
	    */
	}; break;
	case Vec: {
	    if (dim() == 1)
		liften(pea(0));
	}
	}

	return b;
    }

    /**
     * Vereinfacht diesen Term. Dabei handelt es sich um die Standardvereinfachung, die
     * nach Erzeugen eines Terms per Konstruktor und nach Berechnungen ausgef&uuml;hrt 
     * wird. Nachdem alle Unterterme rekursiv vereinfacht wurden, wird u.a. folgendes
     * geleistet: (f(x) bezeichne einen bel. Term, a, b, seien Konstanten)
     * <ul>
     *  <li>Konstanten werden weitestgehend zusammengefa&szlig;t: <tt>1/3 + 1/4 -> 7/12</tt></li>
     *  <li><tt>f(x) * 1 -> f(x), f(x) * 0 -> 0, f(x) + 0 -> f(x)</tt> usw.</li>
     *  <li><tt>a * f(x) + b * f(x) -> (a + b) * f(x)</tt></li>
     *  <li><tt>f(x)^a * f(x)^b -> f(x)^(a + b)</tt></li>
     *  <li><tt>f(x)^(-a) -> 1/f(x)^a</tt>, falls <tt>a</tt> positiv und reell</li>
     *  <li>Skalarprodukte: <tt>(x, y) * (a, b) -> x * a + y * b</tt></li>
     *  <li>Bei <tt>f(x)</tt> selbst wird in der Funktionsliste nachgesehen, ob eine Funktion
     *  namens <tt>f</tt> bekannt ist. Falls ja, wird die simplify-Methode von <tt>f</tt> 
     *  aufgerufen.</li>
     * </ul>
     * @exception ArithmeticException z.B. bei Division durch null oder beim Versuch, mehr als
     * zwei Vektoren miteinander zu multiplizieren.
     */
    public void simplify() {

	while (mySimplify());
	// evtl. existierende Klammern auf hoechster Ebene loeschen
	// also etwa (a * (b+c)) -> a * (b+c)
	if (status == F && name.equals(""))
	    liften(pea(0));
	// Summen mit konst. Faktor auf oberster Ebene ausmultiplizieren:
	// also etwa 2*(a + b) -> 2*a + 2*b
	if (status == S && !faktor.equals(eins)) {
	    for (int i = 0; i < param.size(); i++) 
		pea(i).faktor = (Complex) pea(i).faktor.mul(faktor);
	    faktor = new Complex(1, 0);
	}
    }

    // #############################################################################
    // ###  Verschiedenes  #########################################################
    // #############################################################################

    /**
     * Gibt die Dimension dieses Audrucks zurueck. 
     * Ist dieser Term eine Konstante, so wird ihm die Dimension eins zugeordent. Handelt es
     * sich um eine Funktion, so wird in der Funktionsliste nachgesehen, welche Dimension
     * der Zielbereich dieser Funktion hat. Falls das unbekannt ist oder dieser Term weder
     * Funktion noch Konstante ist, wird 0 zur&uuml;ckgegeben.
     * @return Dimension dieses Terms
     */
    public int dim() {
	
	if (status == K)
	    return 1;
	/*else if (status == F && !name.equals("")) {
	    Funktion f = Funktionsliste.holeFunktion(name);
	    if (f != null)
		return f.dimWertebereich();
		} */
	else if (status == Vec)
	    return param.size();
	
	return 0;
    }
    
    /**
     * Erzeugt einen neuen Term aus diesem Term.
     * @return eine Kopie dieses Terms.
     */
    public Term copy() {

	Term t = 
	    new Term(status, new Complex(faktor.getRealPart(), faktor.getImaginaryPart()), name);
	Enumeration e = param.elements();
	while (e.hasMoreElements())
	    t.param.addElement(((Term) e.nextElement()).copy());
	e = nenner.elements();
	while (e.hasMoreElements())
	    t.nenner.addElement(((Term) e.nextElement()).copy());

	return t;
    }

    /* liefert den i-ten Eintrag des Vectors param (bzw. nenner)*/
    private Term pea(int i) {return (Term) param.elementAt(i);}
    private Term nea(int i) {return (Term) nenner.elementAt(i);}

    /** liefert this, ist redundant, hatte aber mal eine Bedeutung */
    private static Term einbetten(Term t) {

	return t;
    }

    /** ersetzt this durch t. dabei werden die vorfaktoren multipliziert */
    private void liften(Term t) {

	status = t.status;	param = t.param;
	nenner = t.nenner;	name = t.name;
	faktor = (Complex) faktor.mul(t.faktor);
    }

    /** 
     * Pr&uuml;ft, ob dieser Term von einer Variable namens s abh&auml;ngt.
     * @param s Name der Variable, nach der gesucht werden soll
     * @return true, falls s in diesem Term vorkommt.
     */
    public boolean haengtAbVon(String s) {return haengtAbVon(s, true);}

    /** 
     * Pr&uuml;ft, ob dieser Term von einer Funktion namens s abh&auml;ngt.
     * @param s Name der Funktion, nach der gesucht werden soll
     * @return true, falls s in diesem Term vorkommt.
     */
    public boolean haengtAbVonF(String s) {return haengtAbVon(s, false);}

    // bei c = true Variablentest, sonst Funktionstest
    private boolean haengtAbVon(String s, boolean c) {
	
	if (status == K)
	    return false;

	if ((c && status == V || !c && status == F) && name.equals(s))
	    return true;

	boolean b = false;
	for (int i = 0; i < param.size(); i++)
	    b |= pea(i).haengtAbVon(s);
	for (int i = 0; i < nenner.size(); i++)
	    b |= nea(i).haengtAbVon(s);

	return b;     
    }

    // #############################################################################
    // ###  Erzeugung von Termen durch Verknuepfung bestehender Terme  #############
    // #############################################################################

    /**
     * Verknuepft this mit t und liefert das Ergebnis. Die Verknuepfung
     * wird durch i festgelegt: 0: a + t, 1: a - t, 2: a * t
     * 3: a / t, 4: a ^ t
     */
    private static Term verknuepfungsterm(int i, Term a, Term t) {

	Term k = new Term(); k.param.addElement(a.copy());
	if (i == 3) k.nenner.addElement(t.copy());
	else k.param.addElement(t.copy());
	switch (i) {
	case 0: k.status = S; break;
	case 1: k.status = S; k.pea(1).faktor = (Complex) k.pea(1).faktor.mul(-1); break;
	case 2: k.status = P; break;
	case 3: k.status = P; break;
	case 4: k.status = E; break;
	};
	k.simplify();
	return k;
    }

    /** @return this + t */
    public Term plus(Term t) {return verknuepfungsterm(0, this, t);}

    /** @return this - t */
    public Term minus(Term t) {return verknuepfungsterm(1, this, t);}

    /** @return this * t */
    public Term mal(Term t) {return verknuepfungsterm(2, this, t);}

    /** @return this / t */
    public Term durch(Term t) {return verknuepfungsterm(3, this, t);}

    /** @return this ^ t */
    public Term hoch(Term t) {return verknuepfungsterm(4, this, t);}

    // #############################################################################
    // ###  Frontend zu Integration und Ableitung  #################################
    // #############################################################################

    // Hier gestrichen, wird eh nicht verwendet.

    // #############################################################################
    // ###  Berechnung von Funktionswerten, Einsetzen  #############################
    // #############################################################################

    /**
     * Liefert die Zahl, die entsteht, wenn man in diesem Term die
     * Variablen wie angegeben belegt. Die zur&uuml;ckgegebene Zahl ist vom Typ
     * unexakt. Falls der Term nicht zu einer Zahl evaluiert werden kann, weil z.B.
     * noch freie Variablen enthalten sind, gibt diese Methode null zur&uuml;ck.
     * @param variablen Feld der Variablen
     * @param ersetzungen Feld der Belegungen
     * @return Ergebnis
     */
    public Complex einsetzen(String [] variablen, Complex [] ersetzungen) {

	if (variablen.length != ersetzungen.length)
	    return null;

	String [] v = new String[variablen.length + 2];
	Term [] e = new Term[variablen.length + 2];
	for (int i = 0; i < variablen.length; i++) {

	    v[i] = variablen[i].toLowerCase(); 
	    e[i] = new Term(K, ersetzungen[i], "");
	}
	v[v.length - 2] = "pi"; 
	e[v.length - 2] = new Term(K, new Complex(Math.PI, 0), "");
	v[v.length - 1] = "e";
	e[v.length - 1] = new Term(K, new Complex(Math.E, 0), "");

	Term t = copy();
	t.myEinsetzen(v, e);
	t.simplify();
	if (t.status == K)
	    return t.faktor;

	return null;
    }

    private void unexaktMachen() {} // hier redundant

    /**
     * Liefert den Term, der entsteht, wenn man jede Variable
     * variable[i] durch ersetzung[i] ersetzt und das Ergebnis
     * vereinfacht. Gibt null zurueck, falls die Laengen der Felder
     * nicht uebereinstimmen.
     * @param variablen Feld der Variablen
     * @param ersetzungen Feld der Ersetzungen
     * @return Ergebnis
     */
    public Term einsetzen(String [] variablen, Term [] ersetzungen) {

	if (variablen.length != ersetzungen.length)
	    return null;

	Term t = copy();
	t.myEinsetzen(variablen, ersetzungen);
	t.simplify();
	return t;
    }

    private void myEinsetzen(String [] v, Term [] e) {

        for (int i = 0; i < param.size(); i++)
	    pea(i).myEinsetzen(v, e);
	for (int i = 0; i < nenner.size(); i++)
	    nea(i).myEinsetzen(v, e);
	
	if (status == V)
	    for (int i = 0; i < v.length; i++)
		if (v[i].equals(name))
		    liften(e[i].copy());
    }

    /**
     * Ersetzt jedes Vorkommen von u in diesem Term durch v. Anschliessend erfolgt
     * keine Vereinfachung. 
     */
    private void substituiere(Term u, Term v) {

	Term t; Complex uf; Complex tf;
	uf = u.faktor; u.faktor = new Complex(1, 0);
	tf = faktor; faktor = new Complex(1, 0);
	if (equals(u)) {
	    liften(v.copy()); faktor = (Complex) faktor.mul((Complex) tf.div(uf));
	    return;
	}
	faktor = tf;

	for (int i = 0; i < param.size(); i++) 
	    pea(i).substituiere(u, v);
	for (int i = 0; i < nenner.size(); i++)
	    nea(i).substituiere(u, v);
    }

    /**
     * Gibt an, ob dieser Term ein Polynom ist. Polynome sind: Konstanten, Variablen, 
     * Variablen hoch Konstanten, Summen von Polynomen.<br>
     * Achtung: x^sin(1) wird nicht als Polynom erkannt, wohl aber x^(1/3).
     * @return ist dieser Term ein Polynom?
     */
    public boolean istPolynom() {

	boolean b = false;
	switch (status) {
	case V: b = true; break;
	case K: b = true; break;
	case E: 
	    Term t = pea(0); Term v = pea(1);
	    if ((v.status == K) && (t.status == V || t.status == K))
		b = true;
	    break;
	case S: 
	    b = true;
	    for (int i = 0; i < param.size(); i++)
		b &= pea(i).istPolynom();
	}
	return b;
    }

    /**
     * Bestimmt ein Mass fuer die Komplexitaet dieses Terms. Konstanten und Variablen 
     * haben das Mass 1. Weiter gelte:
     * <ul>
     *  <li>tiefe(u + v) = tiefe((u, v)) = max(tiefe(u), tiefe(v)),</li>
     *  <li>tiefe(u * v) = tiefe(u ^ v) =  tiefe(u) + tiefe(v),</li>
     *  <li>tiefe(f(u)) =  1 + tiefe(u)</li>
     * </ul>
     * @return Komplexit&auml;t dieses Terms
     */
    public int tiefe() {

	int t = 0;
	switch (status) {
	case K:
	case V: return 1;
	case F: return 1 + pea(0).tiefe();
	case Vec:
	case S: // maximum
	    for (int i = 0; i < param.size(); i++) {
		int u = pea(i).tiefe();
		if (u > t) t = u;
	    }
	    return t;
	case E:
	case P: // Summe
	    for (int i = 0; i < param.size(); i++) 
		t += pea(i).tiefe();
	    for (int i = 0; i < nenner.size(); i++) 
		t += nea(i).tiefe();
	    return t;
	}

	return 1;
    }

    /**
     * gibt einen Vector zurueck, der alle in diesem Term enthaltenen Variablen 
     * als Strings zurueckgibt.
     */
    public Vector holeVariablen() {
	
	Vector v = new Vector();
	holeVariablen(v);
	return v;
    }

    private void holeVariablen(Vector v) {

	for (int i = 0; i < param.size(); i++)
	    pea(i).holeVariablen(v);
	for (int i = 0; i < nenner.size(); i++)
	    nea(i).holeVariablen(v);

	if (status == V) {
	    boolean gibtsSchon = false;
	    Enumeration e = v.elements();
	    while (e.hasMoreElements())
		gibtsSchon |= name.equals((String)e.nextElement());
	    if (!gibtsSchon)
		v.addElement(name);
	}
    }

    // #############################################################################
    // ###  statische Methoden zum parsen.  ########################################
    // #############################################################################

    /** 
     * Gibt das naechste Vorkommen des uebergebenen chars auf gleicher 
     * Klammerebene an
     * @param w: Quellwort
     * @param i: Startindex 
     * @param c: Zeichen.
     * @return naechstes Vorkommen von c in w nach der i-ten Stelle, 
     * Integer.MAX_VALUE, falls kein Vorkommen mehr ex.
     */
    private static int next(String w, int i, char c) {

	int ebene = 0; char b;
	for (int j = i; j < w.length(); j++) {
	    b = w.charAt(j);
	    if (b == '(' || b == '[' || b == '{')
		ebene++;
	    else if (b == ')' || b == ']' || b == '}')
		ebene--;
	    else if (b == c && ebene == 0) 
		return j;
	}

	return Integer.MAX_VALUE; // nix gefunden
    }

    /** gibt true zurueck, falls c eine Ziffer ist */
    private static boolean istZiffer(char c) {

	if (c == '0' || c == '1' || c == '2' || c == '3' || c == '4' || 
	    c == '5' || c == '6' || c == '7' || c == '8' || c == '9')
	    return true;

	return false;	    
    }
    
    /** gibt true zurueck, falls b ein Buchstabe ist */
    private static boolean istBuchstabe(char b) {

	if (b == 'a' || b == 'f' || b == 'k' || b == 'p' || b == 'u' || 
	    b == 'b' || b == 'g' || b == 'l' || b == 'q' || b == 'v' || 
	    b == 'c' || b == 'h' || b == 'm' || b == 'r' || b == 'w' || 
	    b == 'd' || b == 'i' || b == 'n' || b == 's' || b == 'x' || 
	    b == 'e' || b == 'j' || b == 'o' || b == 't' || b == 'y' || 
	    b == 'z' ||
	    b == 'A' || b == 'F' || b == 'K' || b == 'P' || b == 'U' || 
	    b == 'B' || b == 'G' || b == 'L' || b == 'Q' || b == 'V' || 
	    b == 'C' || b == 'H' || b == 'M' || b == 'R' || b == 'W' || 
	    b == 'D' || b == 'I' || b == 'N' || b == 'S' || b == 'X' || 
	    b == 'E' || b == 'J' || b == 'O' || b == 'T' || b == 'Y' || 
	    b == 'Z')
	    return true;
	return false;
    }

    /** 
     * Bereinigt den String von Leerzeichen, prueft die Syntax und
     * konvertiert in Kleinbuchstaben.
     * Ist die Syntax nicht korrekt, wird eine IllegalArgumentException
     * ausgeworfen.
     */
    private static String ohneLeerzeichen(String s) throws IllegalArgumentException {

	// Leerzeichenbereinigung
	String k = "";
	for (int i = 0; i < s.length(); i++)
	    if (s.charAt(i) != ' ')
		k += s.charAt(i);

	boolean ok = true;
	// Klammerebenen pruefen
	int ebene = 0;
	for (int i = 0; i < k.length(); i++)
	    if (k.charAt(i) == '(')
		ebene++;
	    else if (k.charAt(i) == ')')
		ebene --;
	if (ebene != 0)
	    ok = false;
	
	if (!ok)
	    throw new IllegalArgumentException();
	
	return k;
    }

    // #############################################################################
    // ###  statische Methoden zur Berechnung von Funktionen auf C  ################
    // #############################################################################

    private static final Complex exp(Complex z) {
	
	double exp = Math.exp(z.getRealPart());
	return new Complex(exp * Math.cos(z.getImaginaryPart()),
			   exp * Math.sin(z.getImaginaryPart()));
    }

    private static final Complex log(Complex c) {

	double abs = c.getAbsolutValue();
	return new Complex(Math.log(abs), c.getPolarAngle());
    }

    private static final Complex sqrt(Complex c) {
 
	// z = r * exp(i phi) => sqrt(z) = sqrt(r) * exp(i phi / 2)
	double abs = Math.sqrt(c.getAbsolutValue());
	double phi = c.getPolarAngle() / 2;
	return new Complex(abs * Math.cos(phi), abs * Math.sin(phi));
    }

    private static final Complex re(Complex c) {

    	return new Complex(c.getRealPart(), 0);
    }

    private static final Complex im(Complex c) {

	return new Complex(c.getImaginaryPart(), 0);
    }

    private static final Complex abs(Complex c) {

	return new Complex(c.getAbsolutValue(), 0);
    }

    private static final Complex sin(Complex c) {

	double expB = Math.exp(c.getImaginaryPart());
	double sinAcoshB = Math.sin(c.getRealPart()) * (expB + 1 / expB) / 2;
	double cosAsinhB = Math.cos(c.getRealPart()) * (expB - 1 / expB) / 2;
	return new Complex(sinAcoshB, cosAsinhB);
    }

    private static final Complex cos(Complex c) {

	double exp = Math.exp(c.getImaginaryPart());
	double cosAcoshB = Math.cos(c.getRealPart()) * (exp + 1/exp) / 2;
	double sinAsinhB = -Math.sin(c.getRealPart()) * (exp - 1/exp) / 2;
	return new Complex(cosAcoshB, sinAsinhB);
    }

    private static final Complex tan(Complex c) {

	return (Complex) sin(c).div(cos(c));
    }
    
    private static final Complex arcsin(Complex c) {

	// arcsin(z) = - i * log(i * z + sqrt(1 - z^2))
	return (Complex) Complex.i.mul(log((Complex) Complex.i.mul(c).add(sqrt((Complex) eins.sub(c.mul(c))))));
    }
    
    private static final Complex arccos(Complex c) {

	return (Complex) (new Complex(Math.PI / 2, 0)).sub(arcsin(c));
    }

    private static final Complex arctan(Complex c) {

	// arctan(z) = (log(1 + i*z) - log(1 - i*z)) / (2*i)
	Complex i2 = new Complex(0, 2);
	return (Complex) (log((Complex) eins.add(Complex.i.mul(c))).sub(log((Complex) eins.sub(Complex.i.mul(c))))).div(i2);
    }

    private static final Complex sinh(Complex c) {

	// sinh(z) = (exp(z) - exp(-z))/2
	c = exp(c);
	return (Complex) c.sub((Complex) eins.div(c)).div(zwei);
    }
    
    private static final Complex cosh(Complex c) {

	c = exp(c);
	return (Complex) c.add((Complex) eins.div(c)).div(zwei);
    }
    
    private static final Complex tanh(Complex c) {

	return (Complex) sinh(c).div(cosh(c));
    }
    
    private static final Complex arcsinh(Complex c) {

	// arcsinh(z) = log(z + sqrt(1 + z^2))
	return log((Complex) c.add(sqrt((Complex) eins.add(c.mul(c)))));
    }

    private static final Complex arccosh(Complex c) {

	// arccosh(z) = log(z + (z+1)*sqrt((z-1).durch(z+1)))
	return (Complex) log((Complex) c.add((c.add(eins)).mul(sqrt((Complex) (c.sub(eins)).div(c.add(eins))))));
    }

    private static final Complex arctanh(Complex c) {

	// arctanh(z) = (log(1 + z) - log(1 - z)) / 2
	return (Complex) (log((Complex) c.add(eins)).sub(log((Complex) eins.sub(c)))).div(zwei);
    }

    /**
     * Falls a, b, c keinen (von Null verschiedenen) Imaginaerteil habe,
     * und a < c < b gilt, liefert diese Methode eins, sonst null.
     */
    private static final Complex rect(Complex a, Complex b, Complex c) {
	
	final double eps = 1E-10;

	if (Math.abs(a.getImaginaryPart()) > eps ||
	    Math.abs(b.getImaginaryPart()) > eps ||
	    Math.abs(c.getImaginaryPart()) > eps)
	    return new Complex(0, 0);

	if (a.getRealPart() < c.getRealPart() && c.getRealPart() < b.getRealPart())
	    return new Complex(1, 0);

	return new Complex(0, 0);
    }

    // #############################################################################
    // ###  Testprogramm  ##########################################################
    // #############################################################################

    public static void mymain(String[] args) {

	String [] t = new String[] {"t"};
	String [] x = new String[] {"x"};
	Term [] trm = new Term[1];
	BufferedReader isr = new BufferedReader(new InputStreamReader(System.in));
	while (true) {
	    try {
		System.out.print("> ");
		String s = isr.readLine();
		if (s.charAt(s.length() - 1) == ';')
		    s = s.substring(0, s.length() - 1);
		if (s.equals("exit") || s.equals("quit"))
		    System.exit(0);
		else if (s.length() > 0) {
		    Term f = new Term(s);
		    s = f.toString();
		    if (s.length() < 80)
			for (int i = 0; i < 40 - s.length() / 2; i++)
			    System.out.print(" ");
		    System.out.println(s);	    
		}
	    } catch (ArithmeticException e) {
		System.out.println(e.getMessage());
	    } catch (IllegalArgumentException e) {
		System.out.println("Fehlerhafte Eingabe");
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
    }

    public static void main(String[] args) {

	System.out.println(new Term(args[0]).toString());
    }
}
